# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0


import os
import threading
import time
import unittest
from unittest import mock

from opentelemetry import trace as trace_api
from opentelemetry.context import Context
from opentelemetry.metrics import NoOpMeterProvider
from opentelemetry.sdk import trace
from opentelemetry.sdk.environment_variables import (
    OTEL_BSP_EXPORT_TIMEOUT,
    OTEL_BSP_MAX_EXPORT_BATCH_SIZE,
    OTEL_BSP_MAX_QUEUE_SIZE,
    OTEL_BSP_SCHEDULE_DELAY,
    OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED,
)
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import InMemoryMetricReader
from opentelemetry.sdk.trace import export
from opentelemetry.sdk.trace.export import logger

# pylint: disable=protected-access


class MySpanExporter(export.SpanExporter):
    """Very simple span exporter used for testing."""

    def __init__(
        self,
        destination,
        max_export_batch_size=None,
        export_timeout_millis=0.0,
        export_event: threading.Event = None,
    ):
        self.destination = destination
        self.max_export_batch_size = max_export_batch_size
        self.is_shutdown = False
        self.export_timeout = export_timeout_millis / 1e3
        self.export_event = export_event

    def export(self, spans: trace.Span) -> export.SpanExportResult:
        if (
            self.max_export_batch_size is not None
            and len(spans) > self.max_export_batch_size
        ):
            raise ValueError("Batch is too big")
        time.sleep(self.export_timeout)
        self.destination.extend(span.name for span in spans)
        if self.export_event:
            self.export_event.set()
        return export.SpanExportResult.SUCCESS

    def shutdown(self):
        self.is_shutdown = True


class TestSimpleSpanProcessor(unittest.TestCase):
    def test_simple_span_processor(self):
        tracer_provider = trace.TracerProvider()
        tracer = tracer_provider.get_tracer(__name__)

        spans_names_list = []

        my_exporter = MySpanExporter(destination=spans_names_list)
        span_processor = export.SimpleSpanProcessor(my_exporter)
        tracer_provider.add_span_processor(span_processor)

        with tracer.start_as_current_span("foo"):
            with tracer.start_as_current_span("bar"):
                with tracer.start_as_current_span("xxx"):
                    pass

        self.assertListEqual(["xxx", "bar", "foo"], spans_names_list)

        span_processor.shutdown()
        self.assertTrue(my_exporter.is_shutdown)

    def test_simple_span_processor_no_context(self):
        """Check that we process spans that are never made active.

        SpanProcessors should act on a span's start and end events whether or
        not it is ever the active span.
        """
        tracer_provider = trace.TracerProvider()
        tracer = tracer_provider.get_tracer(__name__)

        spans_names_list = []

        my_exporter = MySpanExporter(destination=spans_names_list)
        span_processor = export.SimpleSpanProcessor(my_exporter)
        tracer_provider.add_span_processor(span_processor)

        with tracer.start_span("foo"):
            with tracer.start_span("bar"):
                with tracer.start_span("xxx"):
                    pass

        self.assertListEqual(["xxx", "bar", "foo"], spans_names_list)

    def test_on_start_accepts_context(self):
        # pylint: disable=no-self-use
        tracer_provider = trace.TracerProvider()
        tracer = tracer_provider.get_tracer(__name__)

        exporter = MySpanExporter([])
        span_processor = mock.Mock(wraps=export.SimpleSpanProcessor(exporter))
        tracer_provider.add_span_processor(span_processor)

        context = Context()
        span = tracer.start_span("foo", context=context)
        span_processor.on_start.assert_called_once_with(
            span, parent_context=context
        )

    def test_simple_span_processor_not_sampled(self):
        tracer_provider = trace.TracerProvider(
            sampler=trace.sampling.ALWAYS_OFF
        )
        tracer = tracer_provider.get_tracer(__name__)

        spans_names_list = []

        my_exporter = MySpanExporter(destination=spans_names_list)
        span_processor = export.SimpleSpanProcessor(my_exporter)
        tracer_provider.add_span_processor(span_processor)

        with tracer.start_as_current_span("foo"):
            with tracer.start_as_current_span("bar"):
                with tracer.start_as_current_span("xxx"):
                    pass

        self.assertListEqual([], spans_names_list)

    @mock.patch.dict(
        "os.environ", {OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED: "true"}
    )
    def test_metrics(self):
        metric_reader = InMemoryMetricReader()
        meter_provider = MeterProvider(metric_readers=[metric_reader])
        num_exports = 0

        def export_spans(_spans):
            nonlocal num_exports
            num_exports += 1
            if num_exports == 3:
                raise RuntimeError("Export failed")
            return export.SpanExportResult.SUCCESS

        exporter = mock.MagicMock()
        exporter.export.side_effect = export_spans
        span_processor = export.SimpleSpanProcessor(
            exporter, meter_provider=meter_provider
        )
        tracer_provider = trace.TracerProvider()
        tracer = tracer_provider.get_tracer(__name__)
        tracer_provider.add_span_processor(span_processor)

        with tracer.start_as_current_span("foo"):
            with tracer.start_span("bar"):
                pass
            with tracer.start_span("baz"):
                pass

        metrics_data = metric_reader.get_metrics_data()
        scope_metrics = metrics_data.resource_metrics[0].scope_metrics[0]
        self.assertEqual(scope_metrics.scope.name, "opentelemetry-sdk")
        metrics = sorted(scope_metrics.metrics, key=lambda m: m.name)
        self.assertEqual(len(metrics), 1)
        self.assertEqual(metrics[0].name, "otel.sdk.processor.span.processed")
        processed_data_points = sorted(
            metrics[0].data.data_points,
            key=lambda dp: dp.attributes.get("error.type", ""),
        )
        self.assertEqual(len(processed_data_points), 2)
        processed_data_point0 = processed_data_points[0]
        self.assertEqual(processed_data_point0.value, 2)
        self.assertEqual(
            processed_data_point0.attributes["otel.component.type"],
            "simple_span_processor",
        )
        self.assertTrue(
            processed_data_point0.attributes["otel.component.name"].startswith(
                "simple_span_processor/"
            )
        )
        self.assertIsNone(processed_data_point0.attributes.get("error.type"))
        processed_data_point1 = processed_data_points[1]
        self.assertEqual(processed_data_point1.value, 1)
        self.assertEqual(
            processed_data_point1.attributes["otel.component.type"],
            "simple_span_processor",
        )
        self.assertTrue(
            processed_data_point1.attributes["otel.component.name"].startswith(
                "simple_span_processor/"
            )
        )
        self.assertEqual(
            processed_data_point1.attributes["error.type"], "RuntimeError"
        )


# Many more test cases for the BatchSpanProcessor exist under
# opentelemetry-sdk/tests/shared_internal/test_batch_processor.py.
# Important: make sure to call .shutdown() on the BatchSpanProcessor
# before the end of the test, otherwise the worker thread will continue
# to run after the end of the test.
class TestBatchSpanProcessor(unittest.TestCase):
    def test_get_span_exporter(self):
        exporter = MySpanExporter(destination=[])
        batch_span_processor = export.BatchSpanProcessor(exporter)
        self.assertEqual(exporter, batch_span_processor.span_exporter)

    @mock.patch.dict(
        "os.environ",
        {
            OTEL_BSP_MAX_QUEUE_SIZE: "10",
            OTEL_BSP_SCHEDULE_DELAY: "2",
            OTEL_BSP_MAX_EXPORT_BATCH_SIZE: "3",
            OTEL_BSP_EXPORT_TIMEOUT: "4",
        },
    )
    def test_args_env_var(self):
        batch_span_processor = export.BatchSpanProcessor(
            MySpanExporter(destination=[])
        )

        self.assertEqual(
            batch_span_processor._batch_processor._max_queue_size, 10
        )
        self.assertEqual(
            batch_span_processor._batch_processor._schedule_delay_millis, 2
        )
        self.assertEqual(
            batch_span_processor._batch_processor._max_export_batch_size, 3
        )
        self.assertEqual(
            batch_span_processor._batch_processor._export_timeout_millis, 4
        )
        batch_span_processor.shutdown()

    def test_args_env_var_defaults(self):
        batch_span_processor = export.BatchSpanProcessor(
            MySpanExporter(destination=[])
        )

        self.assertEqual(
            batch_span_processor._batch_processor._max_queue_size, 2048
        )
        self.assertEqual(
            batch_span_processor._batch_processor._schedule_delay_millis, 5000
        )
        self.assertEqual(
            batch_span_processor._batch_processor._max_export_batch_size, 512
        )
        self.assertEqual(
            batch_span_processor._batch_processor._export_timeout_millis, 30000
        )
        batch_span_processor.shutdown()

    @mock.patch.dict(
        "os.environ",
        {
            OTEL_BSP_MAX_QUEUE_SIZE: "a",
            OTEL_BSP_SCHEDULE_DELAY: " ",
            OTEL_BSP_MAX_EXPORT_BATCH_SIZE: "One",
            OTEL_BSP_EXPORT_TIMEOUT: "@",
        },
    )
    def test_args_env_var_value_error(self):
        logger.disabled = True
        batch_span_processor = export.BatchSpanProcessor(
            MySpanExporter(destination=[])
        )
        logger.disabled = False

        self.assertEqual(
            batch_span_processor._batch_processor._max_queue_size, 2048
        )
        self.assertEqual(
            batch_span_processor._batch_processor._schedule_delay_millis, 5000
        )
        self.assertEqual(
            batch_span_processor._batch_processor._max_export_batch_size, 512
        )
        self.assertEqual(
            batch_span_processor._batch_processor._export_timeout_millis, 30000
        )
        batch_span_processor.shutdown()

    def test_on_start_accepts_parent_context(self):
        # pylint: disable=no-self-use
        my_exporter = MySpanExporter(destination=[])
        span_processor = mock.Mock(
            wraps=export.BatchSpanProcessor(my_exporter)
        )
        tracer_provider = trace.TracerProvider()
        tracer_provider.add_span_processor(span_processor)
        tracer = tracer_provider.get_tracer(__name__)

        context = Context()
        span = tracer.start_span("foo", context=context)

        span_processor.on_start.assert_called_once_with(
            span, parent_context=context
        )

    def test_batch_span_processor_not_sampled(self):
        tracer_provider = trace.TracerProvider(
            sampler=trace.sampling.ALWAYS_OFF
        )
        tracer = tracer_provider.get_tracer(__name__)
        spans_names_list = []

        my_exporter = MySpanExporter(
            destination=spans_names_list, max_export_batch_size=128
        )
        span_processor = export.BatchSpanProcessor(
            my_exporter,
            max_queue_size=256,
            max_export_batch_size=64,
            schedule_delay_millis=100,
        )
        tracer_provider.add_span_processor(span_processor)
        with tracer.start_as_current_span("foo"):
            pass
        time.sleep(0.05)  # give some time for the exporter to upload spans

        span_processor.force_flush()
        self.assertEqual(len(spans_names_list), 0)
        span_processor.shutdown()

    def test_batch_span_processor_parameters(self):
        # zero max_queue_size
        self.assertRaises(
            ValueError, export.BatchSpanProcessor, None, max_queue_size=0
        )

        # negative max_queue_size
        self.assertRaises(
            ValueError,
            export.BatchSpanProcessor,
            None,
            max_queue_size=-500,
        )

        # zero schedule_delay_millis
        self.assertRaises(
            ValueError,
            export.BatchSpanProcessor,
            None,
            schedule_delay_millis=0,
        )

        # negative schedule_delay_millis
        self.assertRaises(
            ValueError,
            export.BatchSpanProcessor,
            None,
            schedule_delay_millis=-500,
        )

        # zero max_export_batch_size
        self.assertRaises(
            ValueError,
            export.BatchSpanProcessor,
            None,
            max_export_batch_size=0,
        )

        # negative max_export_batch_size
        self.assertRaises(
            ValueError,
            export.BatchSpanProcessor,
            None,
            max_export_batch_size=-500,
        )

        # max_export_batch_size > max_queue_size:
        self.assertRaises(
            ValueError,
            export.BatchSpanProcessor,
            None,
            max_queue_size=256,
            max_export_batch_size=512,
        )

    @mock.patch.dict(
        "os.environ", {OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED: "true"}
    )
    def test_metrics(self):  # pylint: disable=too-many-locals,too-many-statements
        metric_reader = InMemoryMetricReader()
        meter_provider = MeterProvider(metric_readers=[metric_reader])
        metric_reader._set_meter_provider(NoOpMeterProvider())
        num_exports = 0
        first_export_event = threading.Event()
        last_export_event = threading.Event()
        run_exports = threading.Event()

        def export_spans(_spans):
            first_export_event.set()
            run_exports.wait()
            nonlocal num_exports
            num_exports += 1
            if num_exports == 3:
                last_export_event.set()
                raise ValueError("Export failed")
            return export.SpanExportResult.SUCCESS

        exporter = mock.MagicMock()
        exporter.export.side_effect = export_spans
        span_processor = export.BatchSpanProcessor(
            exporter,
            meter_provider=meter_provider,
            max_queue_size=1,
            max_export_batch_size=1,
            schedule_delay_millis=1_000_000_000,  # Manually flush
        )
        provider = trace.TracerProvider()
        tracer = provider.get_tracer(__name__)
        provider.add_span_processor(span_processor)

        with tracer.start_as_current_span("foo"):
            pass
        # Wait for span to be sent to exporter
        first_export_event.wait()

        # Queue empty, export in progress, this span is queued
        with tracer.start_as_current_span("bar"):
            pass
        # Queue full, this span is dropped
        with tracer.start_as_current_span("baz"):
            pass

        metrics_data = metric_reader.get_metrics_data()
        scope_metrics = metrics_data.resource_metrics[0].scope_metrics[0]
        self.assertEqual(scope_metrics.scope.name, "opentelemetry-sdk")
        metrics = sorted(scope_metrics.metrics, key=lambda m: m.name)
        self.assertEqual(len(metrics), 3)
        self.assertEqual(metrics[0].name, "otel.sdk.processor.span.processed")
        processed_data_points = sorted(
            metrics[0].data.data_points,
            key=lambda dp: dp.attributes.get("error.type", ""),
        )
        self.assertEqual(len(processed_data_points), 1)
        processed_data_point0 = processed_data_points[0]
        self.assertEqual(processed_data_point0.value, 1)
        self.assertEqual(
            processed_data_point0.attributes["otel.component.type"],
            "batching_span_processor",
        )
        self.assertTrue(
            processed_data_point0.attributes["otel.component.name"].startswith(
                "batching_span_processor/"
            )
        )
        self.assertEqual(
            processed_data_point0.attributes.get("error.type"), "queue_full"
        )
        self.assertEqual(
            metrics[1].name, "otel.sdk.processor.span.queue.capacity"
        )
        queue_capacity_data_point = metrics[1].data.data_points[0]
        self.assertEqual(queue_capacity_data_point.value, 1)
        self.assertEqual(
            queue_capacity_data_point.attributes["otel.component.type"],
            "batching_span_processor",
        )
        self.assertTrue(
            queue_capacity_data_point.attributes[
                "otel.component.name"
            ].startswith("batching_span_processor/")
        )
        self.assertEqual(metrics[2].name, "otel.sdk.processor.span.queue.size")
        queue_size_data_point = metrics[2].data.data_points[0]
        self.assertEqual(queue_size_data_point.value, 1)
        self.assertEqual(
            queue_size_data_point.attributes["otel.component.type"],
            "batching_span_processor",
        )
        self.assertTrue(
            queue_size_data_point.attributes["otel.component.name"].startswith(
                "batching_span_processor/"
            )
        )

        run_exports.set()
        provider.force_flush()

        # This span is processed and exporter returns an error
        with tracer.start_as_current_span("failed"):
            pass
        provider.force_flush()
        last_export_event.wait()

        metrics_data = metric_reader.get_metrics_data()
        scope_metrics = metrics_data.resource_metrics[0].scope_metrics[0]
        self.assertEqual(scope_metrics.scope.name, "opentelemetry-sdk")
        metrics = sorted(scope_metrics.metrics, key=lambda m: m.name)
        self.assertEqual(len(metrics), 3)
        self.assertEqual(metrics[0].name, "otel.sdk.processor.span.processed")
        processed_data_points = sorted(
            metrics[0].data.data_points,
            key=lambda dp: dp.attributes.get("error.type", ""),
        )
        self.assertEqual(len(processed_data_points), 3)
        processed_data_point0 = processed_data_points[0]
        self.assertEqual(processed_data_point0.value, 2)
        self.assertEqual(
            processed_data_point0.attributes["otel.component.type"],
            "batching_span_processor",
        )
        self.assertTrue(
            processed_data_point0.attributes["otel.component.name"].startswith(
                "batching_span_processor/"
            )
        )
        self.assertIsNone(processed_data_point0.attributes.get("error.type"))
        processed_data_point1 = processed_data_points[1]
        self.assertEqual(processed_data_point1.value, 1)
        self.assertEqual(
            processed_data_point1.attributes["otel.component.type"],
            "batching_span_processor",
        )
        self.assertTrue(
            processed_data_point1.attributes["otel.component.name"].startswith(
                "batching_span_processor/"
            )
        )
        self.assertEqual(
            processed_data_point1.attributes.get("error.type"), "ValueError"
        )
        processed_data_point2 = processed_data_points[2]
        self.assertEqual(processed_data_point2.value, 1)
        self.assertEqual(
            processed_data_point2.attributes["otel.component.type"],
            "batching_span_processor",
        )
        self.assertTrue(
            processed_data_point2.attributes["otel.component.name"].startswith(
                "batching_span_processor/"
            )
        )
        self.assertEqual(
            processed_data_point2.attributes.get("error.type"), "queue_full"
        )
        self.assertEqual(
            metrics[1].name, "otel.sdk.processor.span.queue.capacity"
        )
        queue_capacity_data_point = metrics[1].data.data_points[0]
        self.assertEqual(queue_capacity_data_point.value, 1)
        self.assertEqual(
            queue_capacity_data_point.attributes["otel.component.type"],
            "batching_span_processor",
        )
        self.assertTrue(
            queue_capacity_data_point.attributes[
                "otel.component.name"
            ].startswith("batching_span_processor/")
        )
        self.assertEqual(metrics[2].name, "otel.sdk.processor.span.queue.size")
        queue_size_data_point = metrics[2].data.data_points[0]
        self.assertEqual(queue_size_data_point.value, 0)
        self.assertEqual(
            queue_size_data_point.attributes["otel.component.type"],
            "batching_span_processor",
        )
        self.assertTrue(
            queue_size_data_point.attributes["otel.component.name"].startswith(
                "batching_span_processor/"
            )
        )

        provider.shutdown()


class TestConsoleSpanExporter(unittest.TestCase):
    def test_export(self):  # pylint: disable=no-self-use
        """Check that the console exporter prints spans."""

        exporter = export.ConsoleSpanExporter()
        # Mocking stdout interferes with debugging and test reporting, mock on
        # the exporter instance instead.
        span = trace._Span("span name", trace_api.INVALID_SPAN_CONTEXT)
        with mock.patch.object(exporter, "out") as mock_stdout:
            exporter.export([span])
        mock_stdout.write.assert_called_once_with(span.to_json() + os.linesep)

        self.assertEqual(mock_stdout.write.call_count, 1)
        self.assertEqual(mock_stdout.flush.call_count, 1)

    def test_export_custom(self):  # pylint: disable=no-self-use
        """Check that console exporter uses custom io, formatter."""
        mock_span_str = mock.Mock(str)

        def formatter(span):  # pylint: disable=unused-argument
            return mock_span_str

        mock_stdout = mock.Mock()
        exporter = export.ConsoleSpanExporter(
            out=mock_stdout, formatter=formatter
        )
        exporter.export([trace._Span("span name", mock.Mock())])
        mock_stdout.write.assert_called_once_with(mock_span_str)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

__all__ = [
    "ComposableSampler",
    "SamplingIntent",
    "composable_always_off",
    "composable_always_on",
    "composable_parent_threshold",
    "composable_rule_based",
    "composable_traceid_ratio_based",
    "composite_sampler",
]


from ._always_off import composable_always_off
from ._always_on import composable_always_on
from ._composable import ComposableSampler, SamplingIntent
from ._parent_threshold import composable_parent_threshold
from ._rule_based import composable_rule_based
from ._sampler import composite_sampler
from ._traceid_ratio import composable_traceid_ratio_based



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from collections.abc import Sequence

from opentelemetry.context import Context
from opentelemetry.sdk.trace.sampling import Decision, Sampler, SamplingResult
from opentelemetry.trace import Link, SpanKind, TraceState
from opentelemetry.util.types import Attributes

from ._composable import ComposableSampler, SamplingIntent
from ._trace_state import OTEL_TRACE_STATE_KEY, OtelTraceState
from ._util import INVALID_THRESHOLD, is_valid_random_value, is_valid_threshold


class _CompositeSampler(Sampler):
    def __init__(self, delegate: ComposableSampler):
        self._delegate = delegate

    def should_sample(
        self,
        parent_context: Context | None,
        trace_id: int,
        name: str,
        kind: SpanKind | None = None,
        attributes: Attributes | None = None,
        links: Sequence[Link] | None = None,
        trace_state: TraceState | None = None,
    ) -> SamplingResult:
        ot_trace_state = OtelTraceState.parse(trace_state)

        intent = self._delegate.sampling_intent(
            parent_context, name, kind, attributes, links, trace_state
        )
        threshold = intent.threshold

        if is_valid_threshold(threshold):
            adjusted_count_correct = intent.threshold_reliable
            if is_valid_random_value(ot_trace_state.random_value):
                randomness = ot_trace_state.random_value
            else:
                # Use last 56 bits of trace_id as randomness
                randomness = trace_id & 0x00FFFFFFFFFFFFFF
            sampled = threshold <= randomness
        else:
            sampled = False
            adjusted_count_correct = False

        decision = Decision.RECORD_AND_SAMPLE if sampled else Decision.DROP
        if sampled and adjusted_count_correct:
            ot_trace_state.threshold = threshold
        else:
            ot_trace_state.threshold = INVALID_THRESHOLD

        return SamplingResult(
            decision,
            intent.attributes,
            _update_trace_state(trace_state, ot_trace_state, intent),
        )

    def get_description(self) -> str:
        return self._delegate.get_description()


def _update_trace_state(
    trace_state: TraceState | None,
    ot_trace_state: OtelTraceState,
    intent: SamplingIntent,
) -> TraceState | None:
    otts = ot_trace_state.serialize()
    if not trace_state:
        if otts:
            return TraceState(((OTEL_TRACE_STATE_KEY, otts),))
        return None
    new_trace_state = intent.update_trace_state(trace_state)
    if otts:
        return new_trace_state.update(OTEL_TRACE_STATE_KEY, otts)
    return new_trace_state


def composite_sampler(delegate: ComposableSampler) -> Sampler:
    """A sampler that uses a a composable sampler to make its decision while
    handling tracestate.

    Args:
        delegate: The composable sampler to use for making sampling decisions.
    """
    return _CompositeSampler(delegate)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import random
import time

from opentelemetry.metrics import get_meter_provider, set_meter_provider
from opentelemetry.sdk.metrics import Counter, MeterProvider
from opentelemetry.sdk.metrics.export import (
    ConsoleMetricExporter,
    PeriodicExportingMetricReader,
)
from opentelemetry.sdk.metrics.view import DropAggregation, View

# Create a view matching the counter instrument `my.counter`
# and configure the view to drop the aggregation.
drop_aggregation_view = View(
    instrument_type=Counter,
    instrument_name="my.counter",
    aggregation=DropAggregation(),
)

exporter = ConsoleMetricExporter()

reader = PeriodicExportingMetricReader(exporter, export_interval_millis=1_000)
provider = MeterProvider(
    metric_readers=[
        reader,
    ],
    views=[
        drop_aggregation_view,
    ],
)
set_meter_provider(provider)

meter = get_meter_provider().get_meter("view-drop-aggregation", "0.1.2")

my_counter = meter.create_counter("my.counter")

while 1:
    my_counter.add(random.randint(1, 10))
    time.sleep(random.random())



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import threading
import time
import unittest
from logging import WARNING
from unittest.mock import MagicMock, Mock, patch

import requests
from requests import Session
from requests.exceptions import ConnectionError
from requests.models import Response

from opentelemetry.exporter.otlp.proto.http import Compression
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    DEFAULT_COMPRESSION,
    DEFAULT_ENDPOINT,
    DEFAULT_TIMEOUT,
    DEFAULT_TRACES_EXPORT_PATH,
    OTLPSpanExporter,
)
from opentelemetry.exporter.otlp.proto.http.version import __version__
from opentelemetry.sdk.environment_variables import (
    _OTEL_PYTHON_EXPORTER_OTLP_HTTP_TRACES_CREDENTIAL_PROVIDER,
    OTEL_EXPORTER_OTLP_CERTIFICATE,
    OTEL_EXPORTER_OTLP_CLIENT_CERTIFICATE,
    OTEL_EXPORTER_OTLP_CLIENT_KEY,
    OTEL_EXPORTER_OTLP_COMPRESSION,
    OTEL_EXPORTER_OTLP_ENDPOINT,
    OTEL_EXPORTER_OTLP_HEADERS,
    OTEL_EXPORTER_OTLP_TIMEOUT,
    OTEL_EXPORTER_OTLP_TRACES_CERTIFICATE,
    OTEL_EXPORTER_OTLP_TRACES_CLIENT_CERTIFICATE,
    OTEL_EXPORTER_OTLP_TRACES_CLIENT_KEY,
    OTEL_EXPORTER_OTLP_TRACES_COMPRESSION,
    OTEL_EXPORTER_OTLP_TRACES_ENDPOINT,
    OTEL_EXPORTER_OTLP_TRACES_HEADERS,
    OTEL_EXPORTER_OTLP_TRACES_TIMEOUT,
    OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED,
)
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import InMemoryMetricReader
from opentelemetry.sdk.trace import _Span
from opentelemetry.sdk.trace.export import SpanExportResult
from opentelemetry.test.mock_test_classes import IterEntryPoint

OS_ENV_ENDPOINT = "os.env.base"
OS_ENV_CERTIFICATE = "os/env/base.crt"
OS_ENV_CLIENT_CERTIFICATE = "os/env/client-cert.pem"
OS_ENV_CLIENT_KEY = "os/env/client-key.pem"
OS_ENV_HEADERS = "envHeader1=val1,envHeader2=val2,User-agent=Overridden"
OS_ENV_TIMEOUT = "30"
BASIC_SPAN = _Span(
    "abc",
    context=Mock(
        **{
            "trace_state": {"a": "b", "c": "d"},
            "span_id": 10217189687419569865,
            "trace_id": 67545097771067222548457157018666467027,
        }
    ),
)


# pylint: disable=protected-access
class TestOTLPSpanExporter(unittest.TestCase):
    def setUp(self):
        self.metric_reader = InMemoryMetricReader()
        self.meter_provider = MeterProvider(
            metric_readers=[self.metric_reader]
        )

    def test_constructor_default(self):
        exporter = OTLPSpanExporter()

        self.assertEqual(
            exporter._endpoint, DEFAULT_ENDPOINT + DEFAULT_TRACES_EXPORT_PATH
        )
        self.assertEqual(exporter._certificate_file, True)
        self.assertEqual(exporter._client_certificate_file, None)
        self.assertEqual(exporter._client_key_file, None)
        self.assertEqual(exporter._timeout, DEFAULT_TIMEOUT)
        self.assertIs(exporter._compression, DEFAULT_COMPRESSION)
        self.assertEqual(exporter._headers, {})
        self.assertIsInstance(exporter._session, requests.Session)
        self.assertIn("User-Agent", exporter._session.headers)
        self.assertEqual(
            exporter._session.headers.get("Content-Type"),
            "application/x-protobuf",
        )
        self.assertEqual(
            exporter._session.headers.get("User-Agent"),
            "OTel-OTLP-Exporter-Python/" + __version__,
        )

    @patch.dict(
        "os.environ",
        {
            OTEL_EXPORTER_OTLP_CERTIFICATE: OS_ENV_CERTIFICATE,
            OTEL_EXPORTER_OTLP_CLIENT_CERTIFICATE: OS_ENV_CLIENT_CERTIFICATE,
            OTEL_EXPORTER_OTLP_CLIENT_KEY: OS_ENV_CLIENT_KEY,
            OTEL_EXPORTER_OTLP_COMPRESSION: Compression.Gzip.value,
            OTEL_EXPORTER_OTLP_ENDPOINT: OS_ENV_ENDPOINT,
            OTEL_EXPORTER_OTLP_HEADERS: OS_ENV_HEADERS,
            OTEL_EXPORTER_OTLP_TIMEOUT: OS_ENV_TIMEOUT,
            OTEL_EXPORTER_OTLP_TRACES_CERTIFICATE: "traces/certificate.env",
            OTEL_EXPORTER_OTLP_TRACES_CLIENT_CERTIFICATE: "traces/client-cert.pem",
            OTEL_EXPORTER_OTLP_TRACES_CLIENT_KEY: "traces/client-key.pem",
            OTEL_EXPORTER_OTLP_TRACES_COMPRESSION: Compression.Deflate.value,
            OTEL_EXPORTER_OTLP_TRACES_ENDPOINT: "https://traces.endpoint.env",
            OTEL_EXPORTER_OTLP_TRACES_HEADERS: "tracesEnv1=val1,tracesEnv2=val2,traceEnv3===val3==,User-agent=TraceUserAgent",
            OTEL_EXPORTER_OTLP_TRACES_TIMEOUT: "40",
            _OTEL_PYTHON_EXPORTER_OTLP_HTTP_TRACES_CREDENTIAL_PROVIDER: "credential_provider",
        },
    )
    @patch("opentelemetry.exporter.otlp.proto.http._common.entry_points")
    def test_exporter_traces_env_take_priority(self, mock_entry_point):
        credential = Session()

        def f():
            return credential

        mock_entry_point.configure_mock(
            return_value=[IterEntryPoint("custom_credential", f)]
        )
        exporter = OTLPSpanExporter()

        self.assertEqual(exporter._endpoint, "https://traces.endpoint.env")
        self.assertEqual(exporter._certificate_file, "traces/certificate.env")
        self.assertEqual(
            exporter._client_certificate_file, "traces/client-cert.pem"
        )
        self.assertEqual(exporter._client_key_file, "traces/client-key.pem")
        self.assertEqual(exporter._timeout, 40)
        self.assertIs(exporter._compression, Compression.Deflate)
        self.assertEqual(
            exporter._headers,
            {
                "tracesenv1": "val1",
                "tracesenv2": "val2",
                "traceenv3": "==val3==",
                "user-agent": "TraceUserAgent",
            },
        )
        self.assertIs(exporter._session, credential)
        self.assertIsInstance(exporter._session, requests.Session)
        self.assertEqual(
            exporter._session.headers.get("Content-Type"),
            "application/x-protobuf",
        )
        self.assertEqual(
            exporter._session.headers.get("User-Agent"),
            "TraceUserAgent",
        )

    @patch.dict(
        "os.environ",
        {
            OTEL_EXPORTER_OTLP_CERTIFICATE: OS_ENV_CERTIFICATE,
            OTEL_EXPORTER_OTLP_CLIENT_CERTIFICATE: OS_ENV_CLIENT_CERTIFICATE,
            OTEL_EXPORTER_OTLP_CLIENT_KEY: OS_ENV_CLIENT_KEY,
            OTEL_EXPORTER_OTLP_COMPRESSION: Compression.Gzip.value,
            OTEL_EXPORTER_OTLP_ENDPOINT: OS_ENV_ENDPOINT,
            OTEL_EXPORTER_OTLP_TRACES_ENDPOINT: "https://traces.endpoint.env",
            OTEL_EXPORTER_OTLP_HEADERS: OS_ENV_HEADERS,
            OTEL_EXPORTER_OTLP_TIMEOUT: OS_ENV_TIMEOUT,
        },
    )
    def test_exporter_constructor_take_priority(self):
        exporter = OTLPSpanExporter(
            endpoint="example.com/1234",
            certificate_file="path/to/service.crt",
            client_key_file="path/to/client-key.pem",
            client_certificate_file="path/to/client-cert.pem",
            headers={"testHeader1": "value1", "testHeader2": "value2"},
            timeout=20,
            compression=Compression.NoCompression,
            session=requests.Session(),
        )

        self.assertEqual(exporter._endpoint, "example.com/1234")
        self.assertEqual(exporter._certificate_file, "path/to/service.crt")
        self.assertEqual(
            exporter._client_certificate_file, "path/to/client-cert.pem"
        )
        self.assertEqual(exporter._client_key_file, "path/to/client-key.pem")
        self.assertEqual(exporter._timeout, 20)
        self.assertIs(exporter._compression, Compression.NoCompression)
        self.assertEqual(
            exporter._headers,
            {"testHeader1": "value1", "testHeader2": "value2"},
        )
        self.assertIsInstance(exporter._session, requests.Session)

    @patch.dict(
        "os.environ",
        {
            OTEL_EXPORTER_OTLP_CERTIFICATE: OS_ENV_CERTIFICATE,
            OTEL_EXPORTER_OTLP_CLIENT_CERTIFICATE: OS_ENV_CLIENT_CERTIFICATE,
            OTEL_EXPORTER_OTLP_CLIENT_KEY: OS_ENV_CLIENT_KEY,
            OTEL_EXPORTER_OTLP_COMPRESSION: Compression.Gzip.value,
            OTEL_EXPORTER_OTLP_HEADERS: OS_ENV_HEADERS,
            OTEL_EXPORTER_OTLP_TIMEOUT: OS_ENV_TIMEOUT,
        },
    )
    def test_exporter_env(self):
        exporter = OTLPSpanExporter()

        self.assertEqual(exporter._certificate_file, OS_ENV_CERTIFICATE)
        self.assertEqual(
            exporter._client_certificate_file, OS_ENV_CLIENT_CERTIFICATE
        )
        self.assertEqual(exporter._client_key_file, OS_ENV_CLIENT_KEY)
        self.assertEqual(exporter._timeout, int(OS_ENV_TIMEOUT))
        self.assertIs(exporter._compression, Compression.Gzip)
        self.assertEqual(
            exporter._headers,
            {
                "envheader1": "val1",
                "envheader2": "val2",
                "user-agent": "Overridden",
            },
        )

    @patch.dict(
        "os.environ",
        {OTEL_EXPORTER_OTLP_ENDPOINT: OS_ENV_ENDPOINT},
    )
    def test_exporter_env_endpoint_without_slash(self):
        exporter = OTLPSpanExporter()

        self.assertEqual(
            exporter._endpoint,
            OS_ENV_ENDPOINT + f"/{DEFAULT_TRACES_EXPORT_PATH}",
        )

    @patch.dict(
        "os.environ",
        {OTEL_EXPORTER_OTLP_ENDPOINT: OS_ENV_ENDPOINT + "/"},
    )
    def test_exporter_env_endpoint_with_slash(self):
        exporter = OTLPSpanExporter()

        self.assertEqual(
            exporter._endpoint,
            OS_ENV_ENDPOINT + f"/{DEFAULT_TRACES_EXPORT_PATH}",
        )

    @patch.dict(
        "os.environ",
        {
            OTEL_EXPORTER_OTLP_HEADERS: "envHeader1=val1,envHeader2=val2,missingValue"
        },
    )
    def test_headers_parse_from_env(self):
        with self.assertLogs(level="WARNING") as cm:
            _ = OTLPSpanExporter()

            self.assertEqual(
                cm.records[0].message,
                (
                    "Header format invalid! Header values in environment "
                    "variables must be URL encoded per the OpenTelemetry "
                    "Protocol Exporter specification or a comma separated "
                    "list of name=value occurrences: missingValue"
                ),
            )

    @patch.object(OTLPSpanExporter, "_export", return_value=Mock(ok=True))
    def test_2xx_status_code(self, mock_otlp_metric_exporter):
        """
        Test that any HTTP 2XX code returns a successful result
        """

        self.assertEqual(
            OTLPSpanExporter().export(MagicMock()), SpanExportResult.SUCCESS
        )

    @patch.dict("os.environ", {}, clear=True)
    @patch.object(OTLPSpanExporter, "_export", return_value=Mock(ok=True))
    def test_exporter_metrics_disabled_by_default(self, _mock_export):
        exporter = OTLPSpanExporter(meter_provider=self.meter_provider)

        self.assertEqual(
            exporter.export([BASIC_SPAN]), SpanExportResult.SUCCESS
        )

        self.assertIsNone(self.metric_reader.get_metrics_data())

    @patch.dict(
        "os.environ", {OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED: " true "}
    )
    @patch.object(Session, "post")
    def test_retry_timeout(self, mock_post):
        exporter = OTLPSpanExporter(
            timeout=1.5, meter_provider=self.meter_provider
        )

        resp = Response()
        resp.status_code = 503
        resp.reason = "UNAVAILABLE"
        mock_post.return_value = resp
        with self.assertLogs(level=WARNING) as warning:
            before = time.time()
            # Set timeout to 1.5 seconds
            self.assertEqual(
                exporter.export([BASIC_SPAN]),
                SpanExportResult.FAILURE,
            )
            after = time.time()
            # First call at time 0, second at time 1, then an early return before the second backoff sleep b/c it would exceed timeout.
            self.assertEqual(mock_post.call_count, 2)
            # There's a +/-20% jitter on each backoff.
            self.assertTrue(0.75 < after - before < 1.25)
            self.assertIn(
                "Transient error UNAVAILABLE encountered while exporting span batch, retrying in",
                warning.records[0].message,
            )

        metrics_data = self.metric_reader.get_metrics_data()
        scope_metrics = metrics_data.resource_metrics[0].scope_metrics[0]
        self.assertEqual(scope_metrics.scope.name, "opentelemetry-sdk")
        metrics = sorted(scope_metrics.metrics, key=lambda m: m.name)
        self.assertEqual(len(metrics), 3)
        self.assertEqual(
            metrics[0].name, "otel.sdk.exporter.operation.duration"
        )
        self.assert_standard_metric_attrs(
            metrics[0].data.data_points[0].attributes
        )
        self.assertNotIn(
            "error.type", metrics[0].data.data_points[0].attributes
        )
        self.assertEqual(
            metrics[0]
            .data.data_points[0]
            .attributes["http.response.status_code"],
            503,
        )
        self.assertEqual(metrics[1].name, "otel.sdk.exporter.span.exported")
        self.assert_standard_metric_attrs(
            metrics[1].data.data_points[0].attributes
        )
        self.assertNotIn(
            "error.type", metrics[1].data.data_points[0].attributes
        )
        self.assertNotIn(
            "http.response.status_code",
            metrics[1].data.data_points[0].attributes,
        )
        self.assertEqual(metrics[2].name, "otel.sdk.exporter.span.inflight")
        self.assert_standard_metric_attrs(
            metrics[2].data.data_points[0].attributes
        )
        self.assertNotIn(
            "error.type", metrics[2].data.data_points[0].attributes
        )
        self.assertNotIn(
            "http.response.status_code",
            metrics[2].data.data_points[0].attributes,
        )

    @patch.object(Session, "post")
    def test_export_no_collector_available_retryable(self, mock_post):
        exporter = OTLPSpanExporter(timeout=1.5)
        msg = "Server not available."
        mock_post.side_effect = ConnectionError(msg)
        with self.assertLogs(level=WARNING) as warning:
            self.assertEqual(
                exporter.export([BASIC_SPAN]),
                SpanExportResult.FAILURE,
            )
            # Check for greater 2 because the request is on each retry
            # done twice at the moment.
            self.assertGreater(mock_post.call_count, 2)
            self.assertIn(
                f"Transient error {msg} encountered while exporting span batch, retrying in",
                warning.records[0].message,
            )

    @patch.dict(
        "os.environ", {OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED: "true"}
    )
    @patch.object(Session, "post")
    def test_export_no_collector_available(self, mock_post):
        exporter = OTLPSpanExporter(
            timeout=1.5, meter_provider=self.meter_provider
        )

        mock_post.side_effect = requests.exceptions.RequestException()
        with self.assertLogs(level=WARNING) as warning:
            self.assertEqual(
                exporter.export([BASIC_SPAN]),
                SpanExportResult.FAILURE,
            )
            self.assertEqual(mock_post.call_count, 1)
            self.assertIn(
                "Failed to export span batch code",
                warning.records[0].message,
            )

        metrics_data = self.metric_reader.get_metrics_data()
        scope_metrics = metrics_data.resource_metrics[0].scope_metrics[0]
        self.assertEqual(scope_metrics.scope.name, "opentelemetry-sdk")
        metrics = sorted(scope_metrics.metrics, key=lambda m: m.name)
        self.assertEqual(len(metrics), 3)
        self.assertEqual(
            metrics[0].name, "otel.sdk.exporter.operation.duration"
        )
        self.assert_standard_metric_attrs(
            metrics[0].data.data_points[0].attributes
        )
        self.assertEqual(
            metrics[0].data.data_points[0].attributes["error.type"],
            "RequestException",
        )
        self.assertNotIn(
            "http.response.status_code",
            metrics[0].data.data_points[0].attributes,
        )
        self.assertEqual(metrics[1].name, "otel.sdk.exporter.span.exported")
        self.assert_standard_metric_attrs(
            metrics[1].data.data_points[0].attributes
        )
        self.assertEqual(
            metrics[1].data.data_points[0].attributes["error.type"],
            "RequestException",
        )
        self.assertNotIn(
            "http.response.status_code",
            metrics[1].data.data_points[0].attributes,
        )
        self.assertEqual(metrics[2].name, "otel.sdk.exporter.span.inflight")
        self.assert_standard_metric_attrs(
            metrics[2].data.data_points[0].attributes
        )
        self.assertNotIn(
            "error.type", metrics[2].data.data_points[0].attributes
        )
        self.assertNotIn(
            "http.response.status_code",
            metrics[2].data.data_points[0].attributes,
        )

    @patch.object(Session, "post")
    def test_timeout_set_correctly(self, mock_post):
        resp = Response()
        resp.status_code = 200

        def export_side_effect(*args, **kwargs):
            # Timeout should be set to something slightly less than 400 milliseconds depending on how much time has passed.
            self.assertAlmostEqual(0.4, kwargs["timeout"], 2)
            return resp

        mock_post.side_effect = export_side_effect
        exporter = OTLPSpanExporter(timeout=0.4)
        exporter.export([BASIC_SPAN])

    @patch.object(Session, "post")
    def test_shutdown_interrupts_retry_backoff(self, mock_post):
        exporter = OTLPSpanExporter(timeout=1.5)

        resp = Response()
        resp.status_code = 503
        resp.reason = "UNAVAILABLE"
        mock_post.return_value = resp
        thread = threading.Thread(target=exporter.export, args=([BASIC_SPAN],))
        with self.assertLogs(level=WARNING) as warning:
            before = time.time()
            thread.start()
            # Wait for the first attempt to fail, then enter a 1 second backoff.
            time.sleep(0.05)
            # Should cause export to wake up and return.
            exporter.shutdown()
            thread.join()
            after = time.time()
            self.assertIn(
                "Transient error UNAVAILABLE encountered while exporting span batch, retrying in",
                warning.records[0].message,
            )
            self.assertIn(
                "Shutdown in progress, aborting retry.",
                warning.records[1].message,
            )

            assert after - before < 0.2

    def assert_standard_metric_attrs(self, attributes):
        self.assertEqual(
            attributes["otel.component.type"], "otlp_http_span_exporter"
        )
        self.assertTrue(
            attributes["otel.component.name"].startswith(
                "otlp_http_span_exporter/"
            )
        )
        self.assertEqual(attributes["server.address"], "localhost")
        self.assertEqual(attributes["server.port"], 4318)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# AUTO-GENERATED from "opentelemetry/proto/profiles/v1development/profiles.proto"
# DO NOT EDIT MANUALLY

from __future__ import annotations

import builtins
import dataclasses
import functools
import typing

_dataclass = functools.partial(dataclasses.dataclass, slots=True)

import opentelemetry.proto_json._json_codec
import opentelemetry.proto_json.common.v1.common
import opentelemetry.proto_json.resource.v1.resource


@typing.final
@_dataclass
class ProfilesDictionary(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message ProfilesDictionary
    """

    mapping_table: builtins.list[Mapping] = dataclasses.field(default_factory=builtins.list)
    location_table: builtins.list[Location] = dataclasses.field(default_factory=builtins.list)
    function_table: builtins.list[Function] = dataclasses.field(default_factory=builtins.list)
    link_table: builtins.list[Link] = dataclasses.field(default_factory=builtins.list)
    string_table: builtins.list[builtins.str] = dataclasses.field(default_factory=builtins.list)
    attribute_table: builtins.list[KeyValueAndUnit] = dataclasses.field(default_factory=builtins.list)
    stack_table: builtins.list[Stack] = dataclasses.field(default_factory=builtins.list)

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.mapping_table:
            _result["mappingTable"] = opentelemetry.proto_json._json_codec.encode_repeated(self.mapping_table, lambda _v: _v.to_dict())
        if self.location_table:
            _result["locationTable"] = opentelemetry.proto_json._json_codec.encode_repeated(self.location_table, lambda _v: _v.to_dict())
        if self.function_table:
            _result["functionTable"] = opentelemetry.proto_json._json_codec.encode_repeated(self.function_table, lambda _v: _v.to_dict())
        if self.link_table:
            _result["linkTable"] = opentelemetry.proto_json._json_codec.encode_repeated(self.link_table, lambda _v: _v.to_dict())
        if self.string_table:
            _result["stringTable"] = self.string_table
        if self.attribute_table:
            _result["attributeTable"] = opentelemetry.proto_json._json_codec.encode_repeated(self.attribute_table, lambda _v: _v.to_dict())
        if self.stack_table:
            _result["stackTable"] = opentelemetry.proto_json._json_codec.encode_repeated(self.stack_table, lambda _v: _v.to_dict())
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "ProfilesDictionary":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            ProfilesDictionary instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("mappingTable")) is not None:
            _args["mapping_table"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: Mapping.from_dict(_v), "mapping_table")
        if (_value := data.get("locationTable")) is not None:
            _args["location_table"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: Location.from_dict(_v), "location_table")
        if (_value := data.get("functionTable")) is not None:
            _args["function_table"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: Function.from_dict(_v), "function_table")
        if (_value := data.get("linkTable")) is not None:
            _args["link_table"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: Link.from_dict(_v), "link_table")
        if (_value := data.get("stringTable")) is not None:
            _args["string_table"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: _v, "string_table")
        if (_value := data.get("attributeTable")) is not None:
            _args["attribute_table"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: KeyValueAndUnit.from_dict(_v), "attribute_table")
        if (_value := data.get("stackTable")) is not None:
            _args["stack_table"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: Stack.from_dict(_v), "stack_table")

        return cls(**_args)


@typing.final
@_dataclass
class ProfilesData(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message ProfilesData
    """

    resource_profiles: builtins.list[ResourceProfiles] = dataclasses.field(default_factory=builtins.list)
    dictionary: ProfilesDictionary | None = None

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.resource_profiles:
            _result["resourceProfiles"] = opentelemetry.proto_json._json_codec.encode_repeated(self.resource_profiles, lambda _v: _v.to_dict())
        if self.dictionary:
            _result["dictionary"] = self.dictionary.to_dict()
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "ProfilesData":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            ProfilesData instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("resourceProfiles")) is not None:
            _args["resource_profiles"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: ResourceProfiles.from_dict(_v), "resource_profiles")
        if (_value := data.get("dictionary")) is not None:
            _args["dictionary"] = ProfilesDictionary.from_dict(_value)

        return cls(**_args)


@typing.final
@_dataclass
class ResourceProfiles(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message ResourceProfiles
    """

    resource: opentelemetry.proto_json.resource.v1.resource.Resource | None = None
    scope_profiles: builtins.list[ScopeProfiles] = dataclasses.field(default_factory=builtins.list)
    schema_url: builtins.str | None = ""

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.resource:
            _result["resource"] = self.resource.to_dict()
        if self.scope_profiles:
            _result["scopeProfiles"] = opentelemetry.proto_json._json_codec.encode_repeated(self.scope_profiles, lambda _v: _v.to_dict())
        if self.schema_url:
            _result["schemaUrl"] = self.schema_url
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "ResourceProfiles":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            ResourceProfiles instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("resource")) is not None:
            _args["resource"] = opentelemetry.proto_json.resource.v1.resource.Resource.from_dict(_value)
        if (_value := data.get("scopeProfiles")) is not None:
            _args["scope_profiles"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: ScopeProfiles.from_dict(_v), "scope_profiles")
        if (_value := data.get("schemaUrl")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.str, "schema_url")
            _args["schema_url"] = _value

        return cls(**_args)


@typing.final
@_dataclass
class ScopeProfiles(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message ScopeProfiles
    """

    scope: opentelemetry.proto_json.common.v1.common.InstrumentationScope | None = None
    profiles: builtins.list[Profile] = dataclasses.field(default_factory=builtins.list)
    schema_url: builtins.str | None = ""

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.scope:
            _result["scope"] = self.scope.to_dict()
        if self.profiles:
            _result["profiles"] = opentelemetry.proto_json._json_codec.encode_repeated(self.profiles, lambda _v: _v.to_dict())
        if self.schema_url:
            _result["schemaUrl"] = self.schema_url
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "ScopeProfiles":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            ScopeProfiles instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("scope")) is not None:
            _args["scope"] = opentelemetry.proto_json.common.v1.common.InstrumentationScope.from_dict(_value)
        if (_value := data.get("profiles")) is not None:
            _args["profiles"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: Profile.from_dict(_v), "profiles")
        if (_value := data.get("schemaUrl")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.str, "schema_url")
            _args["schema_url"] = _value

        return cls(**_args)


@typing.final
@_dataclass
class Profile(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message Profile
    """

    sample_type: ValueType | None = None
    samples: builtins.list[Sample] = dataclasses.field(default_factory=builtins.list)
    time_unix_nano: builtins.int | None = 0
    duration_nano: builtins.int | None = 0
    period_type: ValueType | None = None
    period: builtins.int | None = 0
    profile_id: builtins.bytes | None = b""
    dropped_attributes_count: builtins.int | None = 0
    original_payload_format: builtins.str | None = ""
    original_payload: builtins.bytes | None = b""
    attribute_indices: builtins.list[builtins.int] = dataclasses.field(default_factory=builtins.list)

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.sample_type:
            _result["sampleType"] = self.sample_type.to_dict()
        if self.samples:
            _result["samples"] = opentelemetry.proto_json._json_codec.encode_repeated(self.samples, lambda _v: _v.to_dict())
        if self.time_unix_nano:
            _result["timeUnixNano"] = opentelemetry.proto_json._json_codec.encode_int64(self.time_unix_nano)
        if self.duration_nano:
            _result["durationNano"] = opentelemetry.proto_json._json_codec.encode_int64(self.duration_nano)
        if self.period_type:
            _result["periodType"] = self.period_type.to_dict()
        if self.period:
            _result["period"] = opentelemetry.proto_json._json_codec.encode_int64(self.period)
        if self.profile_id:
            _result["profileId"] = opentelemetry.proto_json._json_codec.encode_base64(self.profile_id)
        if self.dropped_attributes_count:
            _result["droppedAttributesCount"] = self.dropped_attributes_count
        if self.original_payload_format:
            _result["originalPayloadFormat"] = self.original_payload_format
        if self.original_payload:
            _result["originalPayload"] = opentelemetry.proto_json._json_codec.encode_base64(self.original_payload)
        if self.attribute_indices:
            _result["attributeIndices"] = self.attribute_indices
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "Profile":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            Profile instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("sampleType")) is not None:
            _args["sample_type"] = ValueType.from_dict(_value)
        if (_value := data.get("samples")) is not None:
            _args["samples"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: Sample.from_dict(_v), "samples")
        if (_value := data.get("timeUnixNano")) is not None:
            _args["time_unix_nano"] = opentelemetry.proto_json._json_codec.decode_int64(_value, "time_unix_nano")
        if (_value := data.get("durationNano")) is not None:
            _args["duration_nano"] = opentelemetry.proto_json._json_codec.decode_int64(_value, "duration_nano")
        if (_value := data.get("periodType")) is not None:
            _args["period_type"] = ValueType.from_dict(_value)
        if (_value := data.get("period")) is not None:
            _args["period"] = opentelemetry.proto_json._json_codec.decode_int64(_value, "period")
        if (_value := data.get("profileId")) is not None:
            _args["profile_id"] = opentelemetry.proto_json._json_codec.decode_base64(_value, "profile_id")
        if (_value := data.get("droppedAttributesCount")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "dropped_attributes_count")
            _args["dropped_attributes_count"] = _value
        if (_value := data.get("originalPayloadFormat")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.str, "original_payload_format")
            _args["original_payload_format"] = _value
        if (_value := data.get("originalPayload")) is not None:
            _args["original_payload"] = opentelemetry.proto_json._json_codec.decode_base64(_value, "original_payload")
        if (_value := data.get("attributeIndices")) is not None:
            _args["attribute_indices"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: _v, "attribute_indices")

        return cls(**_args)


@typing.final
@_dataclass
class Link(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message Link
    """

    trace_id: builtins.bytes | None = b""
    span_id: builtins.bytes | None = b""

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.trace_id:
            _result["traceId"] = opentelemetry.proto_json._json_codec.encode_hex(self.trace_id)
        if self.span_id:
            _result["spanId"] = opentelemetry.proto_json._json_codec.encode_hex(self.span_id)
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "Link":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            Link instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("traceId")) is not None:
            _args["trace_id"] = opentelemetry.proto_json._json_codec.decode_hex(_value, "trace_id")
        if (_value := data.get("spanId")) is not None:
            _args["span_id"] = opentelemetry.proto_json._json_codec.decode_hex(_value, "span_id")

        return cls(**_args)


@typing.final
@_dataclass
class ValueType(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message ValueType
    """

    type_strindex: builtins.int | None = 0
    unit_strindex: builtins.int | None = 0

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.type_strindex:
            _result["typeStrindex"] = self.type_strindex
        if self.unit_strindex:
            _result["unitStrindex"] = self.unit_strindex
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "ValueType":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            ValueType instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("typeStrindex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "type_strindex")
            _args["type_strindex"] = _value
        if (_value := data.get("unitStrindex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "unit_strindex")
            _args["unit_strindex"] = _value

        return cls(**_args)


@typing.final
@_dataclass
class Sample(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message Sample
    """

    stack_index: builtins.int | None = 0
    attribute_indices: builtins.list[builtins.int] = dataclasses.field(default_factory=builtins.list)
    link_index: builtins.int | None = 0
    values: builtins.list[builtins.int] = dataclasses.field(default_factory=builtins.list)
    timestamps_unix_nano: builtins.list[builtins.int] = dataclasses.field(default_factory=builtins.list)

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.stack_index:
            _result["stackIndex"] = self.stack_index
        if self.attribute_indices:
            _result["attributeIndices"] = self.attribute_indices
        if self.link_index:
            _result["linkIndex"] = self.link_index
        if self.values:
            _result["values"] = opentelemetry.proto_json._json_codec.encode_repeated(self.values, lambda _v: opentelemetry.proto_json._json_codec.encode_int64(_v))
        if self.timestamps_unix_nano:
            _result["timestampsUnixNano"] = opentelemetry.proto_json._json_codec.encode_repeated(self.timestamps_unix_nano, lambda _v: opentelemetry.proto_json._json_codec.encode_int64(_v))
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "Sample":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            Sample instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("stackIndex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "stack_index")
            _args["stack_index"] = _value
        if (_value := data.get("attributeIndices")) is not None:
            _args["attribute_indices"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: _v, "attribute_indices")
        if (_value := data.get("linkIndex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "link_index")
            _args["link_index"] = _value
        if (_value := data.get("values")) is not None:
            _args["values"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: opentelemetry.proto_json._json_codec.decode_int64(_v, "values"), "values")
        if (_value := data.get("timestampsUnixNano")) is not None:
            _args["timestamps_unix_nano"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: opentelemetry.proto_json._json_codec.decode_int64(_v, "timestamps_unix_nano"), "timestamps_unix_nano")

        return cls(**_args)


@typing.final
@_dataclass
class Mapping(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message Mapping
    """

    memory_start: builtins.int | None = 0
    memory_limit: builtins.int | None = 0
    file_offset: builtins.int | None = 0
    filename_strindex: builtins.int | None = 0
    attribute_indices: builtins.list[builtins.int] = dataclasses.field(default_factory=builtins.list)

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.memory_start:
            _result["memoryStart"] = opentelemetry.proto_json._json_codec.encode_int64(self.memory_start)
        if self.memory_limit:
            _result["memoryLimit"] = opentelemetry.proto_json._json_codec.encode_int64(self.memory_limit)
        if self.file_offset:
            _result["fileOffset"] = opentelemetry.proto_json._json_codec.encode_int64(self.file_offset)
        if self.filename_strindex:
            _result["filenameStrindex"] = self.filename_strindex
        if self.attribute_indices:
            _result["attributeIndices"] = self.attribute_indices
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "Mapping":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            Mapping instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("memoryStart")) is not None:
            _args["memory_start"] = opentelemetry.proto_json._json_codec.decode_int64(_value, "memory_start")
        if (_value := data.get("memoryLimit")) is not None:
            _args["memory_limit"] = opentelemetry.proto_json._json_codec.decode_int64(_value, "memory_limit")
        if (_value := data.get("fileOffset")) is not None:
            _args["file_offset"] = opentelemetry.proto_json._json_codec.decode_int64(_value, "file_offset")
        if (_value := data.get("filenameStrindex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "filename_strindex")
            _args["filename_strindex"] = _value
        if (_value := data.get("attributeIndices")) is not None:
            _args["attribute_indices"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: _v, "attribute_indices")

        return cls(**_args)


@typing.final
@_dataclass
class Stack(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message Stack
    """

    location_indices: builtins.list[builtins.int] = dataclasses.field(default_factory=builtins.list)

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.location_indices:
            _result["locationIndices"] = self.location_indices
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "Stack":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            Stack instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("locationIndices")) is not None:
            _args["location_indices"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: _v, "location_indices")

        return cls(**_args)


@typing.final
@_dataclass
class Location(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message Location
    """

    mapping_index: builtins.int | None = 0
    address: builtins.int | None = 0
    lines: builtins.list[Line] = dataclasses.field(default_factory=builtins.list)
    attribute_indices: builtins.list[builtins.int] = dataclasses.field(default_factory=builtins.list)

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.mapping_index:
            _result["mappingIndex"] = self.mapping_index
        if self.address:
            _result["address"] = opentelemetry.proto_json._json_codec.encode_int64(self.address)
        if self.lines:
            _result["lines"] = opentelemetry.proto_json._json_codec.encode_repeated(self.lines, lambda _v: _v.to_dict())
        if self.attribute_indices:
            _result["attributeIndices"] = self.attribute_indices
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "Location":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            Location instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("mappingIndex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "mapping_index")
            _args["mapping_index"] = _value
        if (_value := data.get("address")) is not None:
            _args["address"] = opentelemetry.proto_json._json_codec.decode_int64(_value, "address")
        if (_value := data.get("lines")) is not None:
            _args["lines"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: Line.from_dict(_v), "lines")
        if (_value := data.get("attributeIndices")) is not None:
            _args["attribute_indices"] = opentelemetry.proto_json._json_codec.decode_repeated(_value, lambda _v: _v, "attribute_indices")

        return cls(**_args)


@typing.final
@_dataclass
class Line(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message Line
    """

    function_index: builtins.int | None = 0
    line: builtins.int | None = 0
    column: builtins.int | None = 0

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.function_index:
            _result["functionIndex"] = self.function_index
        if self.line:
            _result["line"] = opentelemetry.proto_json._json_codec.encode_int64(self.line)
        if self.column:
            _result["column"] = opentelemetry.proto_json._json_codec.encode_int64(self.column)
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "Line":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            Line instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("functionIndex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "function_index")
            _args["function_index"] = _value
        if (_value := data.get("line")) is not None:
            _args["line"] = opentelemetry.proto_json._json_codec.decode_int64(_value, "line")
        if (_value := data.get("column")) is not None:
            _args["column"] = opentelemetry.proto_json._json_codec.decode_int64(_value, "column")

        return cls(**_args)


@typing.final
@_dataclass
class Function(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message Function
    """

    name_strindex: builtins.int | None = 0
    system_name_strindex: builtins.int | None = 0
    filename_strindex: builtins.int | None = 0
    start_line: builtins.int | None = 0

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.name_strindex:
            _result["nameStrindex"] = self.name_strindex
        if self.system_name_strindex:
            _result["systemNameStrindex"] = self.system_name_strindex
        if self.filename_strindex:
            _result["filenameStrindex"] = self.filename_strindex
        if self.start_line:
            _result["startLine"] = opentelemetry.proto_json._json_codec.encode_int64(self.start_line)
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "Function":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            Function instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("nameStrindex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "name_strindex")
            _args["name_strindex"] = _value
        if (_value := data.get("systemNameStrindex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "system_name_strindex")
            _args["system_name_strindex"] = _value
        if (_value := data.get("filenameStrindex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "filename_strindex")
            _args["filename_strindex"] = _value
        if (_value := data.get("startLine")) is not None:
            _args["start_line"] = opentelemetry.proto_json._json_codec.decode_int64(_value, "start_line")

        return cls(**_args)


@typing.final
@_dataclass
class KeyValueAndUnit(opentelemetry.proto_json._json_codec.JsonMessage):
    """
    Generated from protobuf message KeyValueAndUnit
    """

    key_strindex: builtins.int | None = 0
    value: opentelemetry.proto_json.common.v1.common.AnyValue | None = None
    unit_strindex: builtins.int | None = 0

    def to_dict(self) -> builtins.dict[builtins.str, typing.Any]:
        """
        Convert this message to a dictionary with lowerCamelCase keys.

        Returns:
            Dictionary representation following OTLP JSON encoding
        """
        _result = {}
        if self.key_strindex:
            _result["keyStrindex"] = self.key_strindex
        if self.value:
            _result["value"] = self.value.to_dict()
        if self.unit_strindex:
            _result["unitStrindex"] = self.unit_strindex
        return _result

    @builtins.classmethod
    def from_dict(cls, data: builtins.dict[builtins.str, typing.Any]) -> "KeyValueAndUnit":
        """
        Create from a dictionary with lowerCamelCase keys.

        Args:
            data: Dictionary representation following OTLP JSON encoding

        Returns:
            KeyValueAndUnit instance
        """
        opentelemetry.proto_json._json_codec.validate_type(data, builtins.dict, "data")
        _args = {}

        if (_value := data.get("keyStrindex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "key_strindex")
            _args["key_strindex"] = _value
        if (_value := data.get("value")) is not None:
            _args["value"] = opentelemetry.proto_json.common.v1.common.AnyValue.from_dict(_value)
        if (_value := data.get("unitStrindex")) is not None:
            opentelemetry.proto_json._json_codec.validate_type(_value, builtins.int, "unit_strindex")
            _args["unit_strindex"] = _value

        return cls(**_args)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

"""
The purpose of this test is to test for backward compatibility with any user-implementable
interfaces as they were originally defined. For example, changes to the MetricExporter ABC must
be made in such a way that existing implementations (outside of this repo) continue to work
when *called* by the SDK.

This does not apply to classes which are not intended to be overridden by the user e.g.  Meter
and PeriodicExportingMetricReader concrete class. Those may freely be modified in a
backward-compatible way for *callers*.

Ideally, we could use pyright for this as well, but SDK is not type checked atm.
"""

from collections.abc import Iterable, Sequence

from opentelemetry.metrics import CallbackOptions, Observation
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics._internal.export import InMemoryMetricReader
from opentelemetry.sdk.metrics.export import (
    Metric,
    MetricExporter,
    MetricExportResult,
    MetricReader,
    PeriodicExportingMetricReader,
)
from opentelemetry.test import TestCase


# Do not change these classes until after major version 1
class OrigMetricExporter(MetricExporter):
    def export(
        self,
        metrics_data: Sequence[Metric],
        timeout_millis: float = 10_000,
        **kwargs,
    ) -> MetricExportResult:
        pass

    def shutdown(self, timeout_millis: float = 30_000, **kwargs) -> None:
        pass

    def force_flush(self, timeout_millis: float = 10_000) -> bool:
        return True


class OrigMetricReader(MetricReader):
    def _receive_metrics(
        self,
        metrics_data: Iterable[Metric],
        timeout_millis: float = 10_000,
        **kwargs,
    ) -> None:
        pass

    def shutdown(self, timeout_millis: float = 30_000, **kwargs) -> None:
        self.collect()


def orig_callback(options: CallbackOptions) -> Iterable[Observation]:
    yield Observation(2)


class TestBackwardCompat(TestCase):
    def test_metric_exporter(self):
        exporter = OrigMetricExporter()
        meter_provider = MeterProvider(
            metric_readers=[PeriodicExportingMetricReader(exporter)]
        )
        # produce some data
        meter_provider.get_meter("foo").create_counter("mycounter").add(12)
        with self.assertNotRaises(Exception):
            meter_provider.shutdown()

    def test_metric_reader(self):
        reader = OrigMetricReader()
        meter_provider = MeterProvider(metric_readers=[reader])
        # produce some data
        meter_provider.get_meter("foo").create_counter("mycounter").add(12)
        with self.assertNotRaises(Exception):
            meter_provider.shutdown()

    def test_observable_callback(self):
        reader = InMemoryMetricReader()
        meter_provider = MeterProvider(metric_readers=[reader])
        # produce some data
        meter_provider.get_meter("foo").create_counter("mycounter").add(12)
        with self.assertNotRaises(Exception):
            metrics_data = reader.get_metrics_data()

        self.assertEqual(len(metrics_data.resource_metrics), 1)
        self.assertEqual(
            len(metrics_data.resource_metrics[0].scope_metrics), 1
        )
        self.assertEqual(
            len(metrics_data.resource_metrics[0].scope_metrics[0].metrics), 1
        )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from ctypes import c_double, c_uint64
from sys import float_info

# IEEE 754 64-bit floating point numbers use 11 bits for the exponent and 52
# bits for the mantissa.
MANTISSA_WIDTH = 52
EXPONENT_WIDTH = 11

# This mask is equivalent to 52 "1" bits (there are 13 hexadecimal 4-bit "f"s
# in the mantissa mask, 13 * 4 == 52) or 0xfffffffffffff in hexadecimal.
MANTISSA_MASK = (1 << MANTISSA_WIDTH) - 1

# There are 11 bits for the exponent, but the exponent values 0 (11 "0"
# bits) and 2047 (11 "1" bits) have special meanings so the exponent range is
# from 1 to 2046. To calculate the exponent value, 1023 (the bias) is
# subtracted from the exponent, so the exponent value range is from -1022 to
# +1023.
EXPONENT_BIAS = (2 ** (EXPONENT_WIDTH - 1)) - 1

# All the exponent mask bits are set to 1 for the 11 exponent bits.
EXPONENT_MASK = ((1 << EXPONENT_WIDTH) - 1) << MANTISSA_WIDTH

# The sign mask has the first bit set to 1 and the rest to 0.
SIGN_MASK = 1 << (EXPONENT_WIDTH + MANTISSA_WIDTH)

# For normal floating point numbers, the exponent can have a value in the
# range [-1022, 1023].
MIN_NORMAL_EXPONENT = -EXPONENT_BIAS + 1
MAX_NORMAL_EXPONENT = EXPONENT_BIAS

# The smallest possible normal value is 2.2250738585072014e-308.
# This value is the result of using the smallest possible number in the
# mantissa, 1.0000000000000000000000000000000000000000000000000000 (52 "0"s in
# the fractional part) and a single "1" in the exponent.
# Finally 1 * (2 ** -1022) = 2.2250738585072014e-308.
MIN_NORMAL_VALUE = float_info.min

# Greatest possible normal value (1.7976931348623157e+308)
# The binary representation of a float in scientific notation uses (for the
# mantissa) one bit for the integer part (which is implicit) and 52 bits for
# the fractional part. Consider a float binary 1.111. It is equal to 1 + 1/2 +
# 1/4 + 1/8. The greatest possible value in the 52-bit binary mantissa would be
# then 1.1111111111111111111111111111111111111111111111111111 (52 "1"s in the
# fractional part) whose decimal value is 1.9999999999999998. Finally,
# 1.9999999999999998 * (2 ** 1023) = 1.7976931348623157e+308.
MAX_NORMAL_VALUE = float_info.max


def get_ieee_754_exponent(value: float) -> int:
    """
    Gets the exponent of the IEEE 754 representation of a float.
    """

    return (
        (
            # This step gives the integer that corresponds to the IEEE 754
            # representation of a float. For example, consider
            # -MAX_NORMAL_VALUE for an example. We choose this value because
            # of its binary representation which makes easy to understand the
            # subsequent operations.
            #
            # c_uint64.from_buffer(c_double(-MAX_NORMAL_VALUE)).value == 18442240474082181119
            # bin(18442240474082181119) == '0b1111111111101111111111111111111111111111111111111111111111111111'
            #
            # The first bit of the previous binary number is the sign bit: 1 (1 means negative, 0 means positive)
            # The next 11 bits are the exponent bits: 11111111110
            # The next 52 bits are the mantissa bits: 1111111111111111111111111111111111111111111111111111
            #
            # This step isolates the exponent bits, turning every bit outside
            # of the exponent field (sign and mantissa bits) to 0.
            c_uint64.from_buffer(c_double(value)).value & EXPONENT_MASK
            # For the example this means:
            # 18442240474082181119 & EXPONENT_MASK == 9214364837600034816
            # bin(9214364837600034816) == '0b111111111100000000000000000000000000000000000000000000000000000'
            # Notice that the previous binary representation does not include
            # leading zeroes, so the sign bit is not included since it is a
            # zero.
        )
        # This step moves the exponent bits to the right, removing the
        # mantissa bits that were set to 0 by the previous step. This
        # leaves the IEEE 754 exponent value, ready for the next step.
        >> MANTISSA_WIDTH
        # For the example this means:
        # 9214364837600034816 >> MANTISSA_WIDTH == 2046
        # bin(2046) == '0b11111111110'
        # As shown above, these are the original 11 bits that correspond to the
        # exponent.
        # This step subtracts the exponent bias from the IEEE 754 value,
        # leaving the actual exponent value.
    ) - EXPONENT_BIAS
    # For the example this means:
    # 2046 - EXPONENT_BIAS == 1023
    # As mentioned in a comment above, the largest value for the exponent is


def get_ieee_754_mantissa(value: float) -> int:
    return (
        c_uint64.from_buffer(c_double(value)).value
        # This step isolates the mantissa bits. There is no need to do any
        # bit shifting as the mantissa bits are already the rightmost field
        # in an IEEE 754 representation.
        & MANTISSA_MASK
    )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0


from opentelemetry.exporter.otlp.proto.common._internal.metrics_encoder import (
    encode_metrics,
)

__all__ = ["encode_metrics"]



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

"""Test script to check given paths for valid README.rst files."""

import argparse
import sys
from pathlib import Path

import readme_renderer.rst


def is_valid_rst(path):
    """Checks if RST can be rendered on PyPI."""
    with open(path, encoding="utf-8") as readme_file:
        markup = readme_file.read()
    return readme_renderer.rst.render(markup, stream=sys.stderr) is not None


def parse_args():
    parser = argparse.ArgumentParser(
        description="Checks README.rst file in path for syntax errors."
    )
    parser.add_argument(
        "paths", nargs="+", help="paths containing a README.rst to test"
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    return parser.parse_args()


def main():
    args = parse_args()
    error = False

    for path in map(Path, args.paths):
        readme = path / "README.rst"
        try:
            if not is_valid_rst(readme):
                error = True
                print("FAILED: RST syntax errors in", readme)
                continue
        except FileNotFoundError:
            error = True
            print("FAILED: README.rst not found in", path)
            continue
        if args.verbose:
            print("PASSED:", readme)

    if error:
        sys.exit(1)
    print("All clear.")


if __name__ == "__main__":
    main()



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import logging
import unittest
from unittest.mock import patch

from opencensus.trace import trace_options, tracestate
from opencensus.trace.blank_span import BlankSpan as OcBlankSpan
from opencensus.trace.link import Link as OcLink
from opencensus.trace.span import SpanKind
from opencensus.trace.span_context import SpanContext
from opencensus.trace.tracer import Tracer as OcTracer
from opencensus.trace.tracers.noop_tracer import NoopTracer as OcNoopTracer

from opentelemetry import context, trace
from opentelemetry.shim.opencensus import install_shim, uninstall_shim
from opentelemetry.shim.opencensus._shim_span import ShimSpan
from opentelemetry.shim.opencensus._shim_tracer import (
    ShimTracer,
    set_oc_span_in_context,
)


class TestShim(unittest.TestCase):
    def setUp(self):
        uninstall_shim()
        install_shim()

    def tearDown(self):
        uninstall_shim()

    def assert_hasattr(self, obj, key):
        self.assertTrue(hasattr(obj, key))

    def test_shim_tracer_wraps_noop_tracer(self):
        oc_tracer = OcTracer()

        self.assertIsInstance(oc_tracer.tracer, ShimTracer)

        # wrapt.ObjectProxy does the magic here. The ShimTracer should look like the real OC
        # NoopTracer.
        self.assertIsInstance(oc_tracer.tracer, OcNoopTracer)
        self.assert_hasattr(oc_tracer.tracer, "finish")
        self.assert_hasattr(oc_tracer.tracer, "span")
        self.assert_hasattr(oc_tracer.tracer, "start_span")
        self.assert_hasattr(oc_tracer.tracer, "end_span")
        self.assert_hasattr(oc_tracer.tracer, "current_span")
        self.assert_hasattr(oc_tracer.tracer, "add_attribute_to_current_span")
        self.assert_hasattr(oc_tracer.tracer, "list_collected_spans")

    def test_shim_tracer_starts_shim_spans(self):
        oc_tracer = OcTracer()
        with oc_tracer.start_span("foo") as span:
            self.assertIsInstance(span, ShimSpan)

    def test_shim_span_wraps_blank_span(self):
        oc_tracer = OcTracer()
        with oc_tracer.start_span("foo") as span:
            # wrapt.ObjectProxy does the magic here. The ShimSpan should look like the real OC
            # BlankSpan.
            self.assertIsInstance(span, OcBlankSpan)

            # members
            self.assert_hasattr(span, "name")
            self.assert_hasattr(span, "parent_span")
            self.assert_hasattr(span, "start_time")
            self.assert_hasattr(span, "end_time")
            self.assert_hasattr(span, "span_id")
            self.assert_hasattr(span, "attributes")
            self.assert_hasattr(span, "stack_trace")
            self.assert_hasattr(span, "annotations")
            self.assert_hasattr(span, "message_events")
            self.assert_hasattr(span, "links")
            self.assert_hasattr(span, "status")
            self.assert_hasattr(span, "same_process_as_parent_span")
            self.assert_hasattr(span, "_child_spans")
            self.assert_hasattr(span, "context_tracer")
            self.assert_hasattr(span, "span_kind")

            # methods
            self.assert_hasattr(span, "on_create")
            self.assert_hasattr(span, "children")
            self.assert_hasattr(span, "span")
            self.assert_hasattr(span, "add_attribute")
            self.assert_hasattr(span, "add_annotation")
            self.assert_hasattr(span, "add_message_event")
            self.assert_hasattr(span, "add_link")
            self.assert_hasattr(span, "set_status")
            self.assert_hasattr(span, "start")
            self.assert_hasattr(span, "finish")
            self.assert_hasattr(span, "__iter__")
            self.assert_hasattr(span, "__enter__")
            self.assert_hasattr(span, "__exit__")

    def test_add_link_logs_a_warning(self):
        oc_tracer = OcTracer()
        with oc_tracer.start_span("foo") as span:
            with self.assertLogs(level=logging.WARNING):
                span.add_link(OcLink("1", "1"))

    def test_set_span_kind_logs_a_warning(self):
        oc_tracer = OcTracer()
        with oc_tracer.start_span("foo") as span:
            with self.assertLogs(level=logging.WARNING):
                span.span_kind = SpanKind.CLIENT

    # pylint: disable=no-self-use,no-member,protected-access
    def test_shim_span_contextmanager_calls_does_not_call_end(self):
        # This was a bug in first implementation where the underlying OTel span.end() was
        # called after span.__exit__ which caused double-ending the span.
        oc_tracer = OcTracer()
        oc_span = oc_tracer.start_span("foo")

        with patch.object(
            oc_span,
            "_self_otel_span",
            wraps=oc_span._self_otel_span,
        ) as spy_otel_span:
            with oc_span:
                pass

        spy_otel_span.end.assert_not_called()

    def test_set_oc_span_in_context_no_span_id(self):
        # This won't create a span ID and is the default behavior if you don't pass a context
        # when creating the Tracer
        ctx = set_oc_span_in_context(SpanContext(), context.get_current())
        self.assertIs(trace.get_current_span(ctx), trace.INVALID_SPAN)

    def test_set_oc_span_in_context_ids(self):
        ctx = set_oc_span_in_context(
            SpanContext(
                trace_id="ace0216bab2b7ba249761dbb19c871b7",
                span_id="1fead89ecf242225",
            ),
            context.get_current(),
        )
        span_ctx = trace.get_current_span(ctx).get_span_context()

        self.assertEqual(
            trace.format_trace_id(span_ctx.trace_id),
            "ace0216bab2b7ba249761dbb19c871b7",
        )
        self.assertEqual(
            trace.format_span_id(span_ctx.span_id), "1fead89ecf242225"
        )

    def test_set_oc_span_in_context_remote(self):
        for is_from_remote in True, False:
            ctx = set_oc_span_in_context(
                SpanContext(
                    trace_id="ace0216bab2b7ba249761dbb19c871b7",
                    span_id="1fead89ecf242225",
                    from_header=is_from_remote,
                ),
                context.get_current(),
            )
            span_ctx = trace.get_current_span(ctx).get_span_context()
            self.assertEqual(span_ctx.is_remote, is_from_remote)

    def test_set_oc_span_in_context_traceoptions(self):
        for oc_trace_options, expect in [
            # Not sampled
            (
                trace_options.TraceOptions("0"),
                trace.TraceFlags(trace.TraceFlags.DEFAULT),
            ),
            # Sampled
            (
                trace_options.TraceOptions("1"),
                trace.TraceFlags(trace.TraceFlags.SAMPLED),
            ),
        ]:
            ctx = set_oc_span_in_context(
                SpanContext(
                    trace_id="ace0216bab2b7ba249761dbb19c871b7",
                    span_id="1fead89ecf242225",
                    trace_options=oc_trace_options,
                ),
                context.get_current(),
            )
            span_ctx = trace.get_current_span(ctx).get_span_context()
            self.assertEqual(span_ctx.trace_flags, expect)

    def test_set_oc_span_in_context_tracestate(self):
        ctx = set_oc_span_in_context(
            SpanContext(
                trace_id="ace0216bab2b7ba249761dbb19c871b7",
                span_id="1fead89ecf242225",
                tracestate=tracestate.Tracestate({"hello": "tracestate"}),
            ),
            context.get_current(),
        )
        span_ctx = trace.get_current_span(ctx).get_span_context()
        self.assertEqual(
            span_ctx.trace_state, trace.TraceState([("hello", "tracestate")])
        )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=protected-access

import json
import os
import subprocess
import sys
import unittest
from pathlib import Path
from unittest.mock import Mock, patch

from opentelemetry._logs import LogRecord, SeverityNumber
from opentelemetry.attributes import BoundedAttributes
from opentelemetry.context import get_current
from opentelemetry.sdk._logs import (
    Logger,
    LoggerProvider,
    ReadableLogRecord,
    ReadWriteLogRecord,
)
from opentelemetry.sdk._logs._internal import (
    NoOpLogger,
    SynchronousMultiLogRecordProcessor,
    _disable_logger_configurator,
    _LoggerConfig,
    _RuleBasedLoggerConfigurator,
    create_logger_metrics,
)
from opentelemetry.sdk.environment_variables import (
    OTEL_EXPERIMENTAL_RESOURCE_DETECTORS,
    OTEL_SDK_DISABLED,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.util.instrumentation import (
    InstrumentationScope,
    _scope_name_matches_glob,
)
from opentelemetry.semconv.attributes import exception_attributes


class TestLoggerProvider(unittest.TestCase):
    def test_resource(self):
        """
        `LoggerProvider` provides a way to allow a `Resource` to be specified.
        """

        logger_provider_0 = LoggerProvider()
        logger_provider_1 = LoggerProvider()

        self.assertEqual(
            logger_provider_0.resource,
            logger_provider_1.resource,
        )
        self.assertIsInstance(logger_provider_0.resource, Resource)
        self.assertIsInstance(logger_provider_1.resource, Resource)

        resource = Resource({"key": "value"})
        self.assertIs(LoggerProvider(resource=resource).resource, resource)

    def test_update_resource(self):
        initial_resource = Resource({"one": "one", "two": "old"})
        updating_resource = Resource({"two": "new", "three": "three"})
        logger_provider = LoggerProvider(resource=initial_resource)
        logger = logger_provider.get_logger("name")
        other_logger = logger_provider.get_logger(
            "other", attributes={"key": "value"}
        )
        processor_mock = Mock()
        logger_provider.add_log_record_processor(processor_mock)

        logger_provider._update_resource(updating_resource)

        self.assertEqual(
            logger_provider.resource.attributes,
            {"one": "one", "two": "new", "three": "three"},
        )
        self.assertIs(logger.resource, logger_provider.resource)
        self.assertIs(other_logger.resource, logger_provider.resource)

        logger.emit(LogRecord(observed_timestamp=0, body="a log line"))
        log_data = processor_mock.on_emit.call_args.args[0]
        self.assertIs(log_data.resource, logger_provider.resource)

        new_logger = logger_provider.get_logger("new")
        self.assertIs(new_logger.resource, logger_provider.resource)

    @unittest.skipUnless(
        hasattr(os, "fork") and hasattr(os, "register_at_fork"),
        "requires os.fork and os.register_at_fork",
    )
    def test_logger_provider_updates_process_dependent_resource_after_fork(
        self,
    ):
        script_path = (
            Path(__file__).parent
            / "scripts"
            / "logger_provider_resource_after_fork.py"
        )

        result = subprocess.run(
            [sys.executable, str(script_path)],
            capture_output=True,
            check=False,
            text=True,
            env={
                **os.environ,
                OTEL_EXPERIMENTAL_RESOURCE_DETECTORS: "process",
            },
        )
        self.assertEqual(
            result.returncode,
            0,
            f"stdout: {result.stdout}\nstderr: {result.stderr}",
        )
        lines = result.stdout.strip().splitlines()
        child_payload = json.loads(lines[0])
        parent_payload = json.loads(lines[1])

        self.assertEqual(
            parent_payload["parent_resource_pid"], parent_payload["parent_pid"]
        )
        self.assertEqual(
            parent_payload["parent_logger_pid"], parent_payload["parent_pid"]
        )
        self.assertEqual(
            parent_payload["parent_resource_pid_after_fork"],
            parent_payload["parent_pid"],
        )
        self.assertEqual(
            parent_payload["parent_logger_pid_after_fork"],
            parent_payload["parent_pid"],
        )

        self.assertNotEqual(
            child_payload["child_pid"], parent_payload["parent_pid"]
        )
        self.assertEqual(
            child_payload["provider_pid"], child_payload["child_pid"]
        )
        self.assertEqual(
            child_payload["cached_logger_pid"], child_payload["child_pid"]
        )
        self.assertEqual(
            child_payload["new_logger_pid"], child_payload["child_pid"]
        )
        self.assertEqual(
            child_payload["exported_resource_pids"],
            [child_payload["child_pid"], child_payload["child_pid"]],
        )
        self.assertEqual(child_payload["log_bodies"], ["cached", "new"])

    def test_get_logger(self):
        """
        `LoggerProvider.get_logger` arguments are used to create an
        `InstrumentationScope` object on the created `Logger`.
        """

        logger = LoggerProvider().get_logger(
            "name",
            version="version",
            schema_url="schema_url",
            attributes={"key": "value"},
        )

        self.assertEqual(logger._instrumentation_scope.name, "name")
        self.assertEqual(logger._instrumentation_scope.version, "version")
        self.assertEqual(
            logger._instrumentation_scope.schema_url, "schema_url"
        )
        self.assertEqual(
            logger._instrumentation_scope.attributes, {"key": "value"}
        )

    @patch.dict("os.environ", {OTEL_SDK_DISABLED: "true"})
    def test_get_logger_with_sdk_disabled(self):
        logger = LoggerProvider().get_logger(Mock())

        self.assertIsInstance(logger, NoOpLogger)

    @patch.object(Resource, "create")
    def test_logger_provider_init(self, resource_patch):
        logger_provider = LoggerProvider()
        resource_patch.assert_called_once()
        self.assertIsNotNone(logger_provider._resource)
        self.assertTrue(
            isinstance(
                logger_provider._multi_log_record_processor,
                SynchronousMultiLogRecordProcessor,
            )
        )
        self.assertIsNotNone(logger_provider._at_exit_handler)

    def test_default_logger_configurator(self):
        provider = LoggerProvider()
        logger = provider.get_logger("module_name", "1.0", "schema_url")
        other_logger = provider.get_logger(
            "other_module_name", "1.0", "schema_url"
        )
        self.assertTrue(logger._is_enabled())
        self.assertTrue(other_logger._is_enabled())

    def test_logger_provider_with_disabled_configurator(self):
        provider = LoggerProvider(
            _logger_configurator=_disable_logger_configurator
        )
        logger = provider.get_logger("test")
        self.assertFalse(logger._is_enabled())

    def test_logger_provider_with_custom_configurator(self):
        def configurator(scope):
            if scope.name == "disabled_logger":
                return _LoggerConfig(is_enabled=False)
            return _LoggerConfig.default()

        provider = LoggerProvider(_logger_configurator=configurator)
        enabled = provider.get_logger("enabled_logger")
        disabled = provider.get_logger("disabled_logger")
        self.assertTrue(enabled._is_enabled())
        self.assertFalse(disabled._is_enabled())

    def test_set_logger_configurator_updates_existing_loggers(self):
        provider = LoggerProvider()
        logger = provider.get_logger("test")
        self.assertTrue(logger._is_enabled())

        provider._set_logger_configurator(
            logger_configurator=_disable_logger_configurator
        )
        self.assertFalse(logger._is_enabled())

    def test_set_logger_configurator_affects_new_loggers(self):
        provider = LoggerProvider()
        provider._set_logger_configurator(
            logger_configurator=_disable_logger_configurator
        )
        logger = provider.get_logger("new_logger")
        self.assertFalse(logger._is_enabled())

    # pylint: disable-next=no-self-use
    def test_disabled_logger_skips_emit(self):
        provider = LoggerProvider(
            _logger_configurator=_disable_logger_configurator
        )
        logger = provider.get_logger("test")
        processor_mock = Mock()
        provider.add_log_record_processor(processor_mock)

        logger.emit(
            LogRecord(observed_timestamp=0, body="should not be emitted")
        )
        processor_mock.on_emit.assert_not_called()

    def test_rule_based_logger_configurator(self):
        rules = [
            (
                _scope_name_matches_glob(glob_pattern="module_name"),
                _LoggerConfig(is_enabled=True),
            ),
            (
                _scope_name_matches_glob(glob_pattern="other_module_name"),
                _LoggerConfig(is_enabled=False),
            ),
        ]
        configurator = _RuleBasedLoggerConfigurator(
            rules=rules, default_config=_LoggerConfig(is_enabled=True)
        )

        provider = LoggerProvider()
        logger = provider.get_logger("module_name", "1.0", "schema_url")
        other_logger = provider.get_logger(
            "other_module_name", "1.0", "schema_url"
        )

        self.assertTrue(logger._is_enabled())
        self.assertTrue(other_logger._is_enabled())

        provider._set_logger_configurator(logger_configurator=configurator)

        self.assertTrue(logger._is_enabled())
        self.assertFalse(other_logger._is_enabled())

    def test_rule_based_logger_configurator_default_when_rules_dont_match(
        self,
    ):
        rules = [
            (
                _scope_name_matches_glob(glob_pattern="module_name"),
                _LoggerConfig(is_enabled=False),
            ),
        ]
        configurator = _RuleBasedLoggerConfigurator(
            rules=rules, default_config=_LoggerConfig(is_enabled=True)
        )

        provider = LoggerProvider()
        logger = provider.get_logger("module_name", "1.0", "schema_url")
        other_logger = provider.get_logger(
            "other_module_name", "1.0", "schema_url"
        )

        self.assertTrue(logger._is_enabled())
        self.assertTrue(other_logger._is_enabled())

        provider._set_logger_configurator(logger_configurator=configurator)

        self.assertFalse(logger._is_enabled())
        self.assertTrue(other_logger._is_enabled())

    def test_rule_based_configurator_first_match_wins(self):
        disabled_config = _LoggerConfig(is_enabled=False)
        enabled_config = _LoggerConfig(is_enabled=True)
        configurator = _RuleBasedLoggerConfigurator(
            rules=[
                (lambda s: s.name == "foo", disabled_config),
                (lambda s: s.name == "foo", enabled_config),
            ],
            default_config=enabled_config,
        )
        scope = InstrumentationScope("foo", "1.0")
        result = configurator(scope)
        self.assertFalse(result.is_enabled)


class TestReadableLogRecord(unittest.TestCase):
    def setUp(self):
        self.log_record = LogRecord(
            timestamp=1234567890,
            observed_timestamp=1234567891,
            body="Test log message",
            attributes={"key": "value"},
            severity_number=SeverityNumber.INFO,
            severity_text="INFO",
        )
        self.resource = Resource({"service.name": "test-service"})
        self.readable_log_record = ReadableLogRecord(
            log_record=self.log_record,
            resource=self.resource,
            instrumentation_scope=None,
        )

    def test_readable_log_record_is_frozen(self):
        """Test that ReadableLogRecord is frozen and cannot be modified."""
        with self.assertRaises((AttributeError, TypeError)):
            self.readable_log_record.log_record = LogRecord(
                timestamp=999, body="Modified"
            )

    def test_readable_log_record_can_read_attributes(self):
        """Test that ReadableLogRecord provides read access to all fields."""
        self.assertEqual(
            self.readable_log_record.log_record.timestamp, 1234567890
        )
        self.assertEqual(
            self.readable_log_record.log_record.body, "Test log message"
        )
        self.assertEqual(
            self.readable_log_record.log_record.attributes["key"], "value"
        )
        self.assertEqual(
            self.readable_log_record.resource.attributes["service.name"],
            "test-service",
        )


class TestLogger(unittest.TestCase):
    @staticmethod
    def _get_logger():
        log_record_processor_mock = Mock()
        logger = Logger(
            resource=Resource.create({}),
            multi_log_record_processor=log_record_processor_mock,
            instrumentation_scope=InstrumentationScope(
                "name",
                "version",
                "schema_url",
                {"an": "attribute"},
            ),
            logger_metrics=create_logger_metrics(Mock(), False),
            _logger_config=_LoggerConfig.default(),
        )
        return logger, log_record_processor_mock

    def test_can_emit_logrecord(self):
        logger, log_record_processor_mock = self._get_logger()
        log_record = LogRecord(
            observed_timestamp=0,
            body="a log line",
        )

        logger.emit(log_record)
        log_record_processor_mock.on_emit.assert_called_once()
        log_data = log_record_processor_mock.on_emit.call_args.args[0]
        self.assertTrue(isinstance(log_data.log_record, LogRecord))
        self.assertTrue(log_data.log_record is log_record)

    def test_can_emit_api_logrecord(self):
        logger, log_record_processor_mock = self._get_logger()
        api_log_record = LogRecord(
            observed_timestamp=0,
            body="a log line",
        )
        logger.emit(api_log_record)
        log_record_processor_mock.on_emit.assert_called_once()
        log_data = log_record_processor_mock.on_emit.call_args.args[0]
        log_record = log_data.log_record
        self.assertTrue(isinstance(log_record, LogRecord))
        self.assertEqual(log_record.timestamp, None)
        self.assertEqual(log_record.observed_timestamp, 0)
        self.assertIsNotNone(log_record.context)
        self.assertEqual(log_record.severity_number, None)
        self.assertEqual(log_record.severity_text, None)
        self.assertEqual(log_record.body, "a log line")
        self.assertEqual(log_record.attributes, {})
        self.assertEqual(log_record.event_name, None)
        self.assertEqual(log_data.resource, logger.resource)

    def test_can_emit_with_keywords_arguments(self):
        logger, log_record_processor_mock = self._get_logger()

        log_record = LogRecord(
            timestamp=100,
            observed_timestamp=101,
            context=get_current(),
            severity_number=SeverityNumber.WARN,
            severity_text="warn",
            body="a body",
            attributes={"some": "attributes"},
            event_name="event_name",
        )
        logger.emit(log_record)
        log_record_processor_mock.on_emit.assert_called_once()
        log_data = log_record_processor_mock.on_emit.call_args.args[0]
        result_log_record = log_data.log_record
        self.assertTrue(isinstance(result_log_record, LogRecord))
        self.assertEqual(result_log_record.timestamp, 100)
        self.assertEqual(result_log_record.observed_timestamp, 101)
        self.assertIsNotNone(result_log_record.context)
        self.assertEqual(
            result_log_record.severity_number, SeverityNumber.WARN
        )
        self.assertEqual(result_log_record.severity_text, "warn")
        self.assertEqual(result_log_record.body, "a body")
        self.assertEqual(result_log_record.attributes, {"some": "attributes"})
        self.assertEqual(result_log_record.event_name, "event_name")
        self.assertEqual(log_data.resource, logger.resource)

    def test_emit_with_exception_adds_attributes(self):
        logger, log_record_processor_mock = self._get_logger()
        exc = ValueError("boom")

        logger.emit(body="a log line", exception=exc)
        log_record_processor_mock.on_emit.assert_called_once()
        log_data = log_record_processor_mock.on_emit.call_args.args[0]
        attributes = dict(log_data.log_record.attributes)
        self.assertEqual(
            attributes[exception_attributes.EXCEPTION_TYPE], "ValueError"
        )
        self.assertEqual(
            attributes[exception_attributes.EXCEPTION_MESSAGE], "boom"
        )
        self.assertIn(
            "ValueError: boom",
            attributes[exception_attributes.EXCEPTION_STACKTRACE],
        )

    def test_emit_with_raised_exception_has_stacktrace(self):
        logger, log_record_processor_mock = self._get_logger()

        try:
            raise ValueError("boom")
        except ValueError as exc:
            logger.emit(body="error", exception=exc)

        log_record_processor_mock.on_emit.assert_called_once()
        log_data = log_record_processor_mock.on_emit.call_args.args[0]
        stacktrace = dict(log_data.log_record.attributes)[
            exception_attributes.EXCEPTION_STACKTRACE
        ]
        self.assertIn("Traceback (most recent call last)", stacktrace)
        self.assertIn("raise ValueError", stacktrace)

    def test_emit_logrecord_exception_preserves_user_attributes(self):
        logger, log_record_processor_mock = self._get_logger()
        exc = ValueError("boom")
        log_record = LogRecord(
            observed_timestamp=0,
            body="a log line",
            attributes={exception_attributes.EXCEPTION_TYPE: "custom"},
            exception=exc,
        )

        logger.emit(log_record)
        log_record_processor_mock.on_emit.assert_called_once()
        log_data = log_record_processor_mock.on_emit.call_args.args[0]
        attributes = dict(log_data.log_record.attributes)
        self.assertEqual(
            attributes[exception_attributes.EXCEPTION_TYPE], "custom"
        )
        self.assertEqual(
            attributes[exception_attributes.EXCEPTION_MESSAGE], "boom"
        )

    def test_emit_logrecord_exception_with_immutable_attributes(self):
        logger, log_record_processor_mock = self._get_logger()
        exc = ValueError("boom")
        original_attributes = BoundedAttributes(
            attributes={"custom": "value"},
            immutable=True,
            extended_attributes=True,
        )
        log_record = LogRecord(
            observed_timestamp=0,
            body="a log line",
            attributes=original_attributes,
            exception=exc,
        )

        logger.emit(log_record)

        self.assertNotIn(
            exception_attributes.EXCEPTION_TYPE, log_record.attributes
        )
        log_record_processor_mock.on_emit.assert_called_once()
        log_data = log_record_processor_mock.on_emit.call_args.args[0]
        attributes = dict(log_data.log_record.attributes)
        self.assertEqual(attributes["custom"], "value")
        self.assertEqual(
            attributes[exception_attributes.EXCEPTION_TYPE], "ValueError"
        )

    def test_emit_readwrite_logrecord_uses_exception(self):
        logger, log_record_processor_mock = self._get_logger()
        exc = RuntimeError("kaput")
        log_record = LogRecord(
            observed_timestamp=0,
            body="a log line",
            exception=exc,
        )
        readwrite = ReadWriteLogRecord(
            log_record=log_record,
            resource=Resource.create({}),
            instrumentation_scope=logger._instrumentation_scope,
        )

        logger.emit(readwrite)
        log_record_processor_mock.on_emit.assert_called_once()
        log_data = log_record_processor_mock.on_emit.call_args.args[0]
        attributes = dict(log_data.log_record.attributes)
        self.assertEqual(
            attributes[exception_attributes.EXCEPTION_TYPE], "RuntimeError"
        )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from concurrent.futures import (  # pylint: disable=no-name-in-module
    ThreadPoolExecutor,
)

# pylint: disable=import-error
from ..otel_ot_shim_tracer import MockTracer
from ..testcase import OpenTelemetryTestCase


class TestThreads(OpenTelemetryTestCase):
    def setUp(self):  # pylint: disable=invalid-name
        self.tracer = MockTracer()
        self.executor = ThreadPoolExecutor(max_workers=3)

    def test_main(self):
        res = self.executor.submit(self.parent_task, "message").result()
        self.assertEqual(res, "message::response")

        spans = self.tracer.finished_spans()
        self.assertEqual(len(spans), 2)
        self.assertNamesEqual(spans, ["child", "parent"])
        self.assertIsChildOf(spans[0], spans[1])

    def parent_task(self, message):
        with self.tracer.start_active_span("parent") as scope:
            fut = self.executor.submit(self.child_task, message, scope.span)
            res = fut.result()

        return res

    def child_task(self, message, span):
        with self.tracer.scope_manager.activate(span, False):
            with self.tracer.start_active_span("child"):
                return f"{message}::response"



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Final

from opentelemetry.metrics import Histogram, Meter

DNS_LOOKUP_DURATION: Final = "dns.lookup.duration"
"""
Measures the time taken to perform a DNS lookup
Instrument: histogram
Unit: s
"""


def create_dns_lookup_duration(meter: Meter) -> Histogram:
    """Measures the time taken to perform a DNS lookup"""
    return meter.create_histogram(
        name=DNS_LOOKUP_DURATION,
        description="Measures the time taken to perform a DNS lookup.",
        unit="s",
    )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

"""
This example shows how to export traces to multiple destinations.
Each BatchSpanProcessor has its own queue and retry logic, so
destinations do not block each other.
"""

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter as GrpcSpanExporter,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as HttpSpanExporter,
)
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    ConsoleSpanExporter,
)

provider = TracerProvider()
trace.set_tracer_provider(provider)

# Destination 1: OTLP over gRPC
grpc_exporter = GrpcSpanExporter(
    endpoint="http://localhost:4317", insecure=True
)
provider.add_span_processor(BatchSpanProcessor(grpc_exporter))

# Destination 2: OTLP over HTTP
http_exporter = HttpSpanExporter(endpoint="http://localhost:4318/v1/traces")
provider.add_span_processor(BatchSpanProcessor(http_exporter))

# Destination 3: Console (for debugging)
provider.add_span_processor(BatchSpanProcessor(ConsoleSpanExporter()))

tracer = trace.get_tracer(__name__)

with tracer.start_as_current_span("example-request"):
    with tracer.start_as_current_span("fetch-data"):
        print("Spans are exported to all three destinations.")

provider.shutdown()



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import asyncio

from opentracing.ext import tags

# pylint: disable=import-error
from ..otel_ot_shim_tracer import MockTracer
from ..testcase import OpenTelemetryTestCase
from ..utils import get_logger, get_one_by_operation_name, stop_loop_when
from .request_handler import RequestHandler

logger = get_logger(__name__)


class Client:
    def __init__(self, request_handler, loop):
        self.request_handler = request_handler
        self.loop = loop

    async def send_task(self, message):
        request_context = {}

        async def before_handler():
            self.request_handler.before_request(message, request_context)

        async def after_handler():
            self.request_handler.after_request(message, request_context)

        await before_handler()
        await after_handler()

        return f"{message}::response"

    def send(self, message):
        return self.send_task(message)

    def send_sync(self, message):
        return self.loop.run_until_complete(self.send_task(message))


class TestAsyncio(OpenTelemetryTestCase):
    """
    There is only one instance of 'RequestHandler' per 'Client'. Methods of
    'RequestHandler' are executed in different Tasks, and no Span propagation
    among them is done automatically.
    Therefore we cannot use current active span and activate span.
    So one issue here is setting correct parent span.
    """

    def setUp(self):
        self.tracer = MockTracer()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.client = Client(RequestHandler(self.tracer), self.loop)

    def tearDown(self):
        self.loop.close()
        super().tearDown()

    def test_two_callbacks(self):
        res_future1 = self.loop.create_task(self.client.send("message1"))
        res_future2 = self.loop.create_task(self.client.send("message2"))

        stop_loop_when(
            self.loop,
            lambda: len(self.tracer.finished_spans()) >= 2,
            timeout=5.0,
        )
        self.loop.run_forever()

        self.assertEqual("message1::response", res_future1.result())
        self.assertEqual("message2::response", res_future2.result())

        spans = self.tracer.finished_spans()
        self.assertEqual(len(spans), 2)

        for span in spans:
            self.assertEqual(
                span.attributes.get(tags.SPAN_KIND, None),
                tags.SPAN_KIND_RPC_CLIENT,
            )

        self.assertNotSameTrace(spans[0], spans[1])
        self.assertIsNone(spans[0].parent)
        self.assertIsNone(spans[1].parent)

    def test_parent_not_picked(self):
        """Active parent should not be picked up by child."""

        async def do_task():
            with self.tracer.start_active_span("parent"):
                response = await self.client.send_task("no_parent")
                self.assertEqual("no_parent::response", response)

        self.loop.run_until_complete(do_task())

        spans = self.tracer.finished_spans()
        self.assertEqual(len(spans), 2)

        child_span = get_one_by_operation_name(spans, "send")
        self.assertIsNotNone(child_span)

        parent_span = get_one_by_operation_name(spans, "parent")
        self.assertIsNotNone(parent_span)

        # Here check that there is no parent-child relation.
        self.assertIsNotChildOf(child_span, parent_span)

    def test_good_solution_to_set_parent(self):
        """Asyncio and contextvars are integrated, in this case it is not needed
        to activate current span by hand.
        """

        async def do_task():
            with self.tracer.start_active_span("parent"):
                # Set ignore_active_span to False indicating that the
                # framework will do it for us.
                req_handler = RequestHandler(
                    self.tracer,
                    ignore_active_span=False,
                )
                client = Client(req_handler, self.loop)
                response = await client.send_task("correct_parent")

                self.assertEqual("correct_parent::response", response)

            # Send second request, now there is no active parent,
            # but it will be set, ups
            response = await client.send_task("wrong_parent")
            self.assertEqual("wrong_parent::response", response)

        self.loop.run_until_complete(do_task())

        spans = self.tracer.finished_spans()
        self.assertEqual(len(spans), 3)

        parent_span = get_one_by_operation_name(spans, "parent")
        self.assertIsNotNone(parent_span)

        spans = [span for span in spans if span != parent_span]
        self.assertIsChildOf(spans[0], parent_span)
        self.assertIsNotChildOf(spans[1], parent_span)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from opentelemetry.sdk.metrics._internal.aggregation import (
    Aggregation,
    DefaultAggregation,
    DropAggregation,
    ExplicitBucketHistogramAggregation,
    ExponentialBucketHistogramAggregation,
    LastValueAggregation,
    SumAggregation,
)
from opentelemetry.sdk.metrics._internal.view import View

__all__ = [
    "Aggregation",
    "DefaultAggregation",
    "DropAggregation",
    "ExplicitBucketHistogramAggregation",
    "ExponentialBucketHistogramAggregation",
    "LastValueAggregation",
    "SumAggregation",
    "View",
]



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

SERVER_ADDRESS: Final = "server.address"
"""
Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.server_attributes.SERVER_ADDRESS`.
"""

SERVER_PORT: Final = "server.port"
"""
Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.server_attributes.SERVER_PORT`.
"""



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Final

from typing_extensions import deprecated

RPC_CONNECT_RPC_ERROR_CODE: Final = "rpc.connect_rpc.error_code"
"""
Deprecated: Replaced by `rpc.response.status_code`.
"""

RPC_CONNECT_RPC_REQUEST_METADATA_TEMPLATE: Final = (
    "rpc.connect_rpc.request.metadata"
)
"""
Deprecated: Replaced by `rpc.request.metadata`.
"""

RPC_CONNECT_RPC_RESPONSE_METADATA_TEMPLATE: Final = (
    "rpc.connect_rpc.response.metadata"
)
"""
Deprecated: Replaced by `rpc.response.metadata`.
"""

RPC_GRPC_REQUEST_METADATA_TEMPLATE: Final = "rpc.grpc.request.metadata"
"""
Deprecated: Replaced by `rpc.request.metadata`.
"""

RPC_GRPC_RESPONSE_METADATA_TEMPLATE: Final = "rpc.grpc.response.metadata"
"""
Deprecated: Replaced by `rpc.response.metadata`.
"""

RPC_GRPC_STATUS_CODE: Final = "rpc.grpc.status_code"
"""
Deprecated: Use string representation of the gRPC status code on the `rpc.response.status_code` attribute.
"""

RPC_JSONRPC_ERROR_CODE: Final = "rpc.jsonrpc.error_code"
"""
Deprecated: Use string representation of the error code on the `rpc.response.status_code` attribute.
"""

RPC_JSONRPC_ERROR_MESSAGE: Final = "rpc.jsonrpc.error_message"
"""
Deprecated: Use the span status description when reporting JSON-RPC spans.
"""

RPC_JSONRPC_REQUEST_ID: Final = "rpc.jsonrpc.request_id"
"""
Deprecated: Replaced by `jsonrpc.request.id`.
"""

RPC_JSONRPC_VERSION: Final = "rpc.jsonrpc.version"
"""
Deprecated: Replaced by `jsonrpc.protocol.version`.
"""

RPC_MESSAGE_COMPRESSED_SIZE: Final = "rpc.message.compressed_size"
"""
Deprecated: Deprecated, no replacement at this time.
"""

RPC_MESSAGE_ID: Final = "rpc.message.id"
"""
Deprecated: Deprecated, no replacement at this time.
"""

RPC_MESSAGE_TYPE: Final = "rpc.message.type"
"""
Deprecated: Deprecated, no replacement at this time.
"""

RPC_MESSAGE_UNCOMPRESSED_SIZE: Final = "rpc.message.uncompressed_size"
"""
Deprecated: Deprecated, no replacement at this time.
"""

RPC_METHOD: Final = "rpc.method"
"""
The fully-qualified logical name of the method from the RPC interface perspective.
Note: The method name MAY have unbounded cardinality in edge or error cases.

Some RPC frameworks or libraries provide a fixed set of recognized methods
for client stubs and server implementations. Instrumentations for such
frameworks MUST set this attribute to the original method name only
when the method is recognized by the framework or library.

When the method is not recognized, for example, when the server receives
a request for a method that is not predefined on the server, or when
instrumentation is not able to reliably detect if the method is predefined,
the attribute MUST be set to `_OTHER`. In such cases, tracing
instrumentations MUST also set `rpc.method_original` attribute to
the original method value.

If the RPC instrumentation could end up converting valid RPC methods to
`_OTHER`, then it SHOULD provide a way to configure the list of recognized
RPC methods.

The `rpc.method` can be different from the name of any implementing
method/function.
The `code.function.name` attribute may be used to record the fully-qualified
method actually executing the call on the server side, or the
RPC client stub method on the client side.
"""

RPC_METHOD_ORIGINAL: Final = "rpc.method_original"
"""
The original name of the method used by the client.
"""

RPC_REQUEST_METADATA_TEMPLATE: Final = "rpc.request.metadata"
"""
RPC request metadata, `<key>` being the normalized RPC metadata key (lowercase), the value being the metadata values.
Note: Instrumentations SHOULD require an explicit configuration of which metadata values are to be captured.
Including all request metadata values can be a security risk - explicit configuration helps avoid leaking sensitive information.

For example, a property `my-custom-key` with value `["1.2.3.4", "1.2.3.5"]` SHOULD be recorded as
`rpc.request.metadata.my-custom-key` attribute with value `["1.2.3.4", "1.2.3.5"]`.
"""

RPC_RESPONSE_METADATA_TEMPLATE: Final = "rpc.response.metadata"
"""
RPC response metadata, `<key>` being the normalized RPC metadata key (lowercase), the value being the metadata values.
Note: Instrumentations SHOULD require an explicit configuration of which metadata values are to be captured.
Including all response metadata values can be a security risk - explicit configuration helps avoid leaking sensitive information.

For example, a property `my-custom-key` with value `["attribute_value"]` SHOULD be recorded as
the `rpc.response.metadata.my-custom-key` attribute with value `["attribute_value"]`.
"""

RPC_RESPONSE_STATUS_CODE: Final = "rpc.response.status_code"
"""
Status code of the RPC returned by the RPC server or generated by the client.
Note: Usually it represents an error code, but may also represent partial success, warning, or differentiate between various types of successful outcomes.
Semantic conventions for individual RPC frameworks SHOULD document what `rpc.response.status_code` means in the context of that system and which values are considered to represent errors.
"""

RPC_SERVICE: Final = "rpc.service"
"""
Deprecated: Value should be included in `rpc.method` which is expected to be a fully-qualified name.
"""

RPC_SYSTEM: Final = "rpc.system"
"""
Deprecated: Replaced by `rpc.system.name`.
"""

RPC_SYSTEM_NAME: Final = "rpc.system.name"
"""
The Remote Procedure Call (RPC) system.
Note: The client and server RPC systems may differ for the same RPC interaction. For example, a client may use Apache Dubbo or Connect RPC to communicate with a server that uses gRPC since both protocols provide compatibility with gRPC.
"""


@deprecated(
    "The attribute rpc.connect_rpc.error_code is deprecated - Replaced by `rpc.response.status_code`"
)
class RpcConnectRpcErrorCodeValues(Enum):
    CANCELLED = "cancelled"
    """cancelled."""
    UNKNOWN = "unknown"
    """unknown."""
    INVALID_ARGUMENT = "invalid_argument"
    """invalid_argument."""
    DEADLINE_EXCEEDED = "deadline_exceeded"
    """deadline_exceeded."""
    NOT_FOUND = "not_found"
    """not_found."""
    ALREADY_EXISTS = "already_exists"
    """already_exists."""
    PERMISSION_DENIED = "permission_denied"
    """permission_denied."""
    RESOURCE_EXHAUSTED = "resource_exhausted"
    """resource_exhausted."""
    FAILED_PRECONDITION = "failed_precondition"
    """failed_precondition."""
    ABORTED = "aborted"
    """aborted."""
    OUT_OF_RANGE = "out_of_range"
    """out_of_range."""
    UNIMPLEMENTED = "unimplemented"
    """unimplemented."""
    INTERNAL = "internal"
    """internal."""
    UNAVAILABLE = "unavailable"
    """unavailable."""
    DATA_LOSS = "data_loss"
    """data_loss."""
    UNAUTHENTICATED = "unauthenticated"
    """unauthenticated."""


@deprecated(
    "The attribute rpc.grpc.status_code is deprecated - Use string representation of the gRPC status code on the `rpc.response.status_code` attribute"
)
class RpcGrpcStatusCodeValues(Enum):
    OK = 0
    """OK."""
    CANCELLED = 1
    """CANCELLED."""
    UNKNOWN = 2
    """UNKNOWN."""
    INVALID_ARGUMENT = 3
    """INVALID_ARGUMENT."""
    DEADLINE_EXCEEDED = 4
    """DEADLINE_EXCEEDED."""
    NOT_FOUND = 5
    """NOT_FOUND."""
    ALREADY_EXISTS = 6
    """ALREADY_EXISTS."""
    PERMISSION_DENIED = 7
    """PERMISSION_DENIED."""
    RESOURCE_EXHAUSTED = 8
    """RESOURCE_EXHAUSTED."""
    FAILED_PRECONDITION = 9
    """FAILED_PRECONDITION."""
    ABORTED = 10
    """ABORTED."""
    OUT_OF_RANGE = 11
    """OUT_OF_RANGE."""
    UNIMPLEMENTED = 12
    """UNIMPLEMENTED."""
    INTERNAL = 13
    """INTERNAL."""
    UNAVAILABLE = 14
    """UNAVAILABLE."""
    DATA_LOSS = 15
    """DATA_LOSS."""
    UNAUTHENTICATED = 16
    """UNAUTHENTICATED."""


@deprecated(
    "The attribute rpc.message.type is deprecated - Deprecated, no replacement at this time"
)
class RpcMessageTypeValues(Enum):
    SENT = "SENT"
    """sent."""
    RECEIVED = "RECEIVED"
    """received."""


@deprecated(
    "The attribute rpc.system is deprecated - Replaced by `rpc.system.name`"
)
class RpcSystemValues(Enum):
    GRPC = "grpc"
    """gRPC."""
    JAVA_RMI = "java_rmi"
    """Java RMI."""
    DOTNET_WCF = "dotnet_wcf"
    """.NET WCF."""
    APACHE_DUBBO = "apache_dubbo"
    """Apache Dubbo."""
    CONNECT_RPC = "connect_rpc"
    """Connect RPC."""
    ONC_RPC = "onc_rpc"
    """[ONC RPC (Sun RPC)](https://datatracker.ietf.org/doc/html/rfc5531)."""
    JSONRPC = "jsonrpc"
    """JSON-RPC."""


class RpcSystemNameValues(Enum):
    GRPC = "grpc"
    """[gRPC](https://grpc.io/)."""
    DUBBO = "dubbo"
    """[Apache Dubbo](https://dubbo.apache.org/)."""
    CONNECTRPC = "connectrpc"
    """[Connect RPC](https://connectrpc.com/)."""
    JSONRPC = "jsonrpc"
    """[JSON-RPC](https://www.jsonrpc.org/)."""
